from PIL.Image import MODES
from ingradient_library.dataloads import *
import matplotlib.pyplot as plt
from tqdm import tqdm
from torchvision.transforms import CenterCrop
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import cc3d
import SimpleITK as sitk
import time
import os

def get_kernel(patch_size, sigma = 1e+3):
    kernel_shape = (patch_size[0], patch_size[1], patch_size[2])
    tmp = tuple([torch.arange(i) for i in kernel_shape])
    gx, gy, gz = torch.meshgrid(*tmp)
    gx = torch.abs(gx - kernel_shape[0]//2)
    gy = torch.abs(gy - kernel_shape[1]//2)
    gz = torch.abs(gz - kernel_shape[2]//2)
    kernel = torch.exp(-((gx**2 + gy**2 + gz**2)/(2*(sigma))))

    return kernel

def get_pad_images(images, patch_size):
    #이미지가 patch 보다 작을 경우 padding
    images = torch.tensor(images)
    if len(patch_size) == 3:
        image_shape = torch.tensor([images.shape[-3], images.shape[-2], images.shape[-1]])
    else:
        image_shape = torch.tensor([images.shape[-2], images.shape[-1]])
    prev_shape = torch.tensor(image_shape)
    patch_size = torch.tensor(patch_size)
    while len(images.shape) < 5:
        images = images.unsqueeze(0)
    
    new_shape = torch.tensor([0,0,0])
    for i in range(len(patch_size)):
        if patch_size[i] > image_shape[i]:
            new_shape[i] = (patch_size[i]) - prev_shape[i]
    is_div_two = new_shape % 2 != 0
    pad_shape = new_shape.repeat_interleave(2) // 2
    for i in range(len(is_div_two)):
        if is_div_two[i]:
            pad_shape[i*2] += 1

    images = F.pad(images, pad_shape.tolist()[::-1], "constant", 0)

    while len(images.shape) > 4:
        images = images.squeeze(0)

    return images

def get_coords_arr(images, patch_size, step_size):
    #image와 patch, step size를 통해 patch를 뽑는 위치를 grid 단위로 나타냄. 
    grid_list = []
    image_shape = torch.tensor([images.shape[-3], images.shape[-2], images.shape[-1]])
    for i in range(3):
        step = 1+(image_shape[i] - patch_size[i])/(patch_size[i] * step_size[i])
        if step < 1:
            step = 1.0
        grid_list.append(torch.arange(step))
    grid_x, grid_y, grid_z = torch.meshgrid(*grid_list)
    grid_x = torch.round(grid_x * step_size[0] * patch_size[0]).long()
    grid_y = torch.round(grid_y * step_size[1] * patch_size[1]).long()
    grid_z = torch.round(grid_z * step_size[2] * patch_size[2]).long()
    max_gx = grid_x.max()
    max_gy = grid_y.max()
    max_gz = grid_z.max()

    grid_x[torch.where(grid_x == max_gx)] = images.shape[-3] - patch_size[0]
    grid_y[torch.where(grid_y == max_gy)] = images.shape[-2] - patch_size[1]
    grid_z[torch.where(grid_z == max_gz)] = images.shape[-1] - patch_size[2]
    grid_x = grid_x.reshape(-1, 1)
    grid_y = grid_y.reshape(-1, 1)
    grid_z = grid_z.reshape(-1, 1)
    
    grids = torch.cat((grid_x, grid_y, grid_z), -1)
    patch_coords = []
    for i in range(grids.shape[0]):
        patch_coords.append((torch.arange(grids[i][0], grids[i][0]  + patch_size[0]), torch.arange(grids[i][1], grids[i][1] + patch_size[1]), torch.arange(grids[i][2], grids[i][2] + patch_size[2]) ) )
    return patch_coords


def get_patch_from_coords(image, patch_coords, index):
    #위의 함수에서 얻어진 image당 patch의 coords에서 image와 coords가 주어졌을 대 patch를 뽑아냄
    for i in range(image.shape[0]):
        if i == 0:
            patch = image[0][torch.meshgrid(*patch_coords[index])].unsqueeze(0)
        else:
            patch = torch.cat((patch, image[i][torch.meshgrid(*patch_coords[index])].unsqueeze(0)))
    return patch

def change_value_from_coords(image, patch_coords, index, patch):
    #patch의 값으로 image와 해당하는 coords 의 값을 바꿈
    for i in range(image.shape[0]):
        image[i][torch.meshgrid(*patch_coords[index])] += patch[i]
    return image
"""

class TTA(object):
    def __init__(self, flip = [0,1,2], permute = [[1,0,2]]):
        self.flip = flip
        self.permute = permute

    def transform(self, image):
        images = [image]
        transformed_log = [] # {'image':, 'flip':, 'permute': 

        if self.permute != None:
            for p in self.permute:
                permutted_img = image.permute(p[0],p[1],p[2])
                image.append(permutted_img)

        for f in self.flip:
            flipped_img = torch.flip(image, f)
            images.append(flipped_img)
            if self.permute != None:
                for p in self.permute:
                    permutted_img = flipped_img.permute(p[0],p[1],p[2])
                    image.append(permutted_img)

        return images
        
    def inverse_transform(self):

"""

class CC_Remover(object):
    # remain num => 남길 connected components label의 수
    # frequency order => connected component labeling 된 것 중 택할 빈도 수를 정함. 1은 보통 background 임.
    def __init__(self, remain_num = 2, frequency_order = [2,3], distance = True, axis_z = 2, cc_label_counts = 1500, distance_weight = 20.0):
        self.remain_num = remain_num
        self.frequency_order = frequency_order
        self.distance = distance
        self.axis_z = axis_z
        self.cc_label_counts = cc_label_counts
        self.distance_weight = distance_weight

    def __call__(self, mask):
        conected_component_mask = cc3d.connected_components(mask, connectivity = 26)
        count = np.unique(conected_component_mask, return_counts = True)[1]
        result = torch.zeros(mask.shape)

        if self.distance != False:
            component_length = (count > self.cc_label_counts).sum()
            remained_components_indices = np.argsort(count)[::-1][:component_length]
            center_coords = []
            min_max_coords = []
            for r_i in remained_components_indices:
                min_max_coords.append([np.where(conected_component_mask == r_i)[0].max() - np.where(conected_component_mask == r_i)[0].min(), np.where(conected_component_mask == r_i)[1].max() - np.where(conected_component_mask == r_i)[1].min(), np.where(conected_component_mask == r_i)[2].max()- np.where(conected_component_mask == r_i)[2].min()])
                center_coords.append([np.where(conected_component_mask == r_i)[self.axis_z].mean()])
            center_coords = np.array(center_coords)
            min_max_coords = np.array(min_max_coords)
            n_clusters = center_coords.shape[0]
            distance = np.zeros((n_clusters, n_clusters))

            for i in range(n_clusters):
                for j in range(n_clusters):
                    if i == j:
                        distance[i][j] = 1e+10
                    else:
                        distance[i][j] = self.distance_weight * numpy.linalg.norm(center_coords[i]-center_coords[j]) + numpy.linalg.norm(min_max_coords[i]-min_max_coords[j])

            x, y = np.unravel_index(np.argmin(distance, axis=None), distance.shape)
            return torch.tensor((remained_components_indices[x] == conected_component_mask) + (remained_components_indices[y] == conected_component_mask))
        else:
            for i in range(self.remain_num):
                try:
                    result += (np.argsort(count)[::-1][self.frequency_order[i]-1] == conected_component_mask)
                except IndexError as e:
                    print(e)
                
            return torch.tensor(result)

class InferenceV2(object):
    """
    dataset : torch.utils.data 의 dataset 또는 해당 class를 상속받은 개체.
    patch_size : inference 시에 적용 할 patch size를 설정
    n_classes : softmax output의 channel 수
    mode : 저장할 데이터셋 포멧 npy와 nifti만 지원 됨.
    deep_supervision : 모델에 deep_supervision이 적용 될 경우, True
    padding : patch size에 맞게 padding
    """
    def __init__(self, dataset, model, patch_size, n_classes, mode = 'npy and nifti', step_size = [0.5, 0.5, 0.5], postprocessing = None, device = None, softmax_threshold = 0.99, tta = None,
                 batch_size = 1, gaussian_filter = True, deep_supervision = False, save_path = None, arg_max = False, padding = False, is_2d = False, sigma = 1e+3):
        self.dataset = dataset
        self.model = model.to(device)
        self.bs = batch_size
        self.patch_size = patch_size
        self.step_size = step_size
        self.n_classes = n_classes
        self.deep_supervision = deep_supervision
        self.device = device
        self.postprocessing = postprocessing
        self.mode = mode
        self.save_path = save_path
        if save_path != None:
            self.make_dir(save_path)
        self.arg_max = arg_max
        self.padding = padding
        self.is_2d = is_2d
        self.softmax_threshold = softmax_threshold
        if gaussian_filter == True:
            self.gaussian_kernel = get_kernel(patch_size, sigma = sigma)
            self.gaussian_kernel = self.gaussian_kernel.to(device)
        
        self.tta = tta
        
    def valid_run(self):
        #전체 데이터에 대해 inference 시작.
        print('start valid run')
        final_score = []
        self.model = self.model.eval()
        for data_index in range(len(self.dataset)):
            if isinstance(self.dataset, CustomDataset):
                current_name = self.dataset.image_name[data_index][:-4]
            else:
                current_name = 'result'
            image = self.dataset[data_index][0]
            seg = self.dataset[data_index][1]
            if self.padding:
                seg = get_pad_images(seg, self.patch_size)
            
            if isinstance(seg, np.ndarray):
                seg = torch.from_numpy(seg)
            
            #if self.tta != None:
            #    input_list = self.tta.transform(image)
            #    output_list = []
            #    for input in input_list:
            result = self.get_result(image, data_index, current_name)

            dice_score = []
            for i in range(1, self.n_classes):
                temp_seg = (seg == i).long()
                temp_img = (result == i).long()
                intersection = (temp_seg * temp_img).sum()
                union = temp_seg.sum() + temp_img.sum()
                dice = 2 * intersection / union
                dice_score.append(dice.numpy())
            print("file_name : ", self.dataset.image_name[data_index], "dice : ", dice_score[0].item())
            final_score.append(dice_score)
        
        self.final_score = final_score
        return final_score

    def make_dir(self, save_path):
        now = time.localtime()
        now = "%04d%02d%02d_%02d%02d%02d" % (now.tm_year, now.tm_mon, now.tm_mday, now.tm_hour, now.tm_min, now.tm_sec)

        self.save_path = os.path.join(save_path, now)
        self.soft_path = os.path.join(self.save_path, r'soft_result')
        self.model_path = os.path.join(self.save_path, r'result')
        self.postprocessing_path = os.path.join(self.save_path, r'postprocessing_result')
        self.image_path = os.path.join(self.save_path, r'image')
        self.nii_img_path = os.path.join(self.save_path, r'nifti_image')

        try:
            os.mkdir(self.save_path)
            os.mkdir(self.soft_path)
            os.mkdir(self.image_path)
            os.mkdir(self.model_path)
            os.mkdir(self.postprocessing_path)
            os.mkdir(self.nii_img_path)
        
        except FileExistsError as e:
            print('Folders already exist. Please be careful to overwrite exising files.')

    def save(self, image, path, index, identifier, mode = 'seg', is_nifti = False):
        if 'npy' in self.mode:
            np.save(os.path.join(path, identifier + mode + '{0:03}.npy'.format(index)),image.detach().cpu().numpy())
        
        if is_nifti:
            nii_img = sitk.GetImageFromArray(image.detach().cpu().numpy())
            sitk.WriteImage(nii_img, os.path.join(self.nii_img_path, identifier + mode + '{0:03}.nii.gz'.format(index)))

    def test_run(self):
        for data_index in range(len(self.dataset)):
            if isinstance(self.dataset, CustomDataset):
                current_name = self.dataset.image_name[data_index][:-4]
            image = self.dataset[data_index][0]
            result = self.get_result(image, data_index, current_name)
            
        return result
    
    def get_result(self, image, data_index = None, current_name = None):
        # 각 이미지 별로 result를 구함.
        if self.padding:
            image = get_pad_images(image, self.patch_size)
        coords = get_coords_arr(image, self.patch_size, self.step_size)
        self.model.eval()

        for coords_index in range(len(coords)):
            if coords_index == 0:
                patch_bundle = get_patch_from_coords(image, coords, coords_index).unsqueeze(0)
                index_bundle = torch.tensor(coords_index).reshape(1,)
            else:
                patch_bundle = torch.cat((patch_bundle, get_patch_from_coords(image, coords, coords_index).unsqueeze(0)), 0)
                index_bundle = torch.cat((index_bundle, torch.tensor(coords_index).reshape(1,)))

        patch_bundle = torch.split(patch_bundle, self.bs)
        index_bundle = torch.split(index_bundle, self.bs)
        result = torch.zeros((self.n_classes, image.shape[-3], image.shape[-2], image.shape[-1]))
        for mini_batch_patch, mini_batch_index in tqdm(zip(patch_bundle, index_bundle)):
            with torch.no_grad():
                output = self.model(mini_batch_patch.to(self.device).float())
                if self.deep_supervision:
                    output = output[0]
                output = torch.softmax(output, dim = 1) * self.gaussian_kernel.unsqueeze(0).unsqueeze(0)
                output = output.detach().cpu()
                for batch_index in range(mini_batch_index.shape[0]):
                    result = change_value_from_coords(result, coords, mini_batch_index[batch_index], output[batch_index])
        
        if self.save_path != None:
            self.save(result, self.soft_path, data_index, current_name)
        
        if self.arg_max:
            result = torch.argmax(result, 0)
        else:
            result = torch.softmax(result, dim = 0)
            result[1:] = result[1:] >= self.softmax_threshold
            result = torch.argmax(result, 0)

        if self.save_path != None:
            argmax_result = torch.argmax(result, 0), 
            self.save(argmax_result, self.model_path, data_index, current_name, mode = 'seg', is_nifti = True)
            self.save(image[0], self.image_path, data_index, current_name, mode = 'image', is_nifti = True)

            if self.postprocessing != None:
                mask = self.postprocessing((argmax_result != 0).detach().cpu().numpy())
                argmax_result = mask * argmax_result
                self.save(argmax_result, self.model_path, data_index, current_name, mode = 'postprocessing', is_nifti = True)
        
        return result
    



class Inference2D(InferenceV2):
    def __init__(self, model, dataset, axis = 2, patch_size = (512, 512), save_path = None, n_classes = 2, padding = True, deepsupervision = False, softmax_threshold = 0.99):
        if isinstance(dataset, Dataset2D):
            dataset.oversampling = False
        
        self.model = model
        self.dataset = dataset
        self.axis = axis
        self.centercrop = CenterCrop(size = patch_size)
        self.save_path = save_path
        self.n_classes = n_classes
        self.padding = padding
        self.patch_size = patch_size
        self.deepsupervision = deepsupervision
        self.softmax_threshold = softmax_threshold
        if save_path != None:
            self.make_dir(save_path)
    
    def save(self, image, path, index, identifier, mode='seg', is_nifti=False):
        return super().save(image, path, index, identifier, mode=mode, is_nifti=is_nifti)
    
    def make_dir(self, save_path):
        return super().make_dir(save_path)

    def get_value_by_index(self, tensor, index):
        if self.axis == 0:
            result = tensor[:, index]
        elif self.axis == 1:
            result = tensor[:, :, index]
        elif self.axis == 2:
            result = tensor[:, :, :, index]
        return result
    
    def save_value_by_index(self, tensor, input, index):
        if self.axis == 0:
            tensor[:, index] = input
        elif self.axis == 1:
            tensor[:, :, index] = input
        elif self.axis == 2:
            tensor[:, :, :, index] = input
        return tensor

    def valid_run(self):
        return super().valid_run()
        
    def get_result(self, image, data_index, current_name):
        result = torch.zeros(image.shape)
        
        if self.axis == 0:
            axis_range = image.shape[-3]
        elif self.axis == 1:
            axis_range = image.shape[-2]
        elif self.axis == 2:
            axis_range = image.shape[-1]

        
        for i in range(axis_range):
            input_img = self.get_value_by_index(image, i) # bs, n_modality, x, y 
            while len(input_img.shape) < 4:
                input_img = input_img.unsqueeze(0)
            
            if self.deepsupervision:
                output = self.model(input_img)[0]
            else:
                output = self.model(input_img)
            result = self.save_value_by_index(result, output, i)
        

        if self.arg_max:
            result = torch.argmax(result, 0)
        else:
            result[:, 1:] = (result[:, 1:] >= self.softmax_threshold)
            result = torch.softmax(result, dim = 1)

        if self.save_path != None:
            self.save(result, self.model_path, data_index, current_name, mode = 'seg', is_nifti = True)
            self.save(image[0], self.image_path, data_index, current_name, mode = 'image', is_nifti = True)
        
        return result


from numpy.random.mtrand import normal, sample
import torch
import torch.nn as nn
import torch.nn.functional as F
import SimpleITK as sitk
from torch.utils.data import Dataset
import numpy as np
import pickle
import numpy as numpy
import os
from ingradient_library.patch_transform import Batch_Affine_3D
from ingradient_library.preprocessing import *
from ingradient_library.transform import *
from torchvision.transforms import CenterCrop
import h5py


def random_axis_sample(images, seg, axis, mode = 'over'):
    if isinstance(images, np.ndarray) or isinstance(seg, np.ndarray):
        images = torch.tensor(images)
        seg = torch.tensor(seg)
    if mode == 'over':
        candidates = torch.where(seg != 0)[axis]
        index = torch.randint(low = candidates.min(), high = candidates.max() + 1, size = (1,)).item()
    
    elif mode == 'random':
        index = torch.randint(low = 0, high = seg.shape[axis], size = (1,)).item()

    if axis == 0:
        images = images[:,index]
        seg = seg[index]
    
    elif axis == 1:
        images = images[:, :, index]
        seg = seg[:, index]
    
    elif axis == 2:
        images = images[:, :, :, index]
        seg = seg[:, :, index]

    return images, seg

def set_direction(img, direction):
    if isinstance(img, np.ndarray):
        img = torch.tensor(img)
    img = img.permute(2,1,0)
    img = img.numpy()
    direction = np.array(direction).reshape(3,3)
    inverse = np.where(direction == -1)[0]
    permute = np.where(direction != 0)[1]
    img = img.transpose(permute)
    if 0 in inverse:
        img = img[::-1, :, :]

    if 1 in inverse:
        img = img[:, ::-1, :]

    if 2 in inverse:
        img = img[:, :, ::-1]

    return torch.tensor(img.copy())


class Dataset2D(Dataset):
    def __init__(self, root_dir, axis, mode, hdf_img_keys = None, train = True, hdf_seg_key = None ,sampler = None, size = (512, 512), gru_size = (512,512,512),no_sampling = False,
                oversampling = True, device = None, normalizer = None, transform = None, resample = None, correct_direction = True):

        self.file_name = []
        self.info_name = []
        for file_name in os.listdir(root_dir):
            if mode == 'npz':
                if 'npz' in file_name and not 'py' in file_name:
                    self.file_name.append(file_name)
            
                elif 'pkl' in file_name and not 'py' in file_name:
                    self.info_name.append(file_name)
            elif mode == 'hdf':
                if not 'py' in file_name and 'hdf' in file_name:
                    self.file_name.append(file_name)
        
        self.root_dir = root_dir
        self.axis = axis
        self.mode = mode
        self.hdf_img_keys = hdf_img_keys
        self.hdf_seg_key = hdf_seg_key
        self.sampler = sampler
        self.oversampling = oversampling
        self.device = device
        self.normalizer = normalizer
        self.transform = transform
        self.resample = resample
        self.correct_direction = correct_direction
        self.train = train
        self.crop = CenterCrop(size = size)
        self.no_sampling = no_sampling
        
    def __len__(self):
        return len(self.file_name)
    
    def resample(self, x, info_data, y = None):
        if self.train:
                if isinstance(y, np.ndarray):
                    y = torch.tensor(y.copy())
                result = self.resample(torch.vstack((x, y.unsqueeze(0))), info_data)
                x = result[:-1]
                y = result[-1].squeeze(0)
                return x, y
            
        else:
            x = self.resample(x, info_data)
            return x, y
    
    def __getitem__(self, idx):
        current_file = os.path.join(self.root_dir, self.file_name[idx])
        
        if self.mode == 'npz':
            info_file = open(os.path.join(self.root_dir,self.info_name[idx]), 'rb')
            info_data = pickle.load(info_file)
            info_file.close()
            data = np.load(os.path.join(self.root_dir, self.file_name[idx]))
            x = data['x']
            y = data['y']

        elif self.mode == 'hdf':
            hdf_file = h5py.File(current_file , 'r')
            first_check = True
            for h_key in self.hdf_img_keys:
                if first_check:
                    x = np.array(hdf_file[h_key])
                    first_check = False
                else:
                    x = np.stack((np.expand_dims(x, axis = 0), np.expand_dims(hdf_file[h_key], axis = 0)))
            y = np.array(hdf_file[self.hdf_seg_key[0]])
        

        if self.correct_direction:
            for i in range(len(x)):
                if i == 0:
                    result = set_direction(x[i], info_data['direction']).unsqueeze(0)
                else:
                    result = torch.vstack((result, set_direction(x[i], info_data['direction']).unsqueeze(0)))
            x = result 
            if self.train:
                y = set_direction(y, info_data['direction'])

        if self.normalizer != None:
            x = self.normalizer(x)
        
        if self.resample != None:
            x, y = self.resample(x, info_data, y)
        
        if self.sampler != None:
            x, y = self.sampler(x, y)
        
        


        if self.oversampling == False:
            return x, y
        if self.oversampling:
            temp = np.random.uniform(0, 1)
            if temp < 0.5:
                x, y = random_axis_sample(x, y, self.axis, mode = 'over')
            else:
                x, y = random_axis_sample(x, y, self.axis, mode = 'random')
        
        else:
            x, y = random_axis_sample(x, y, self.axis, mode = 'random')
    
        x = self.crop(x)
        y = self.crop(y)
        return x, y

        

class CustomDataset(Dataset):
    def __init__(self, root_dir, device = None, normalizer = None, mode = 'train', dim = '3d', correct_direction = True, resample = None, save_file_list = True):
        self.root_dir = root_dir
        self.image_name = []
        self.info_name = []
        self.device = device
        self.normalizer = normalizer
        self.mode = mode
        self.dim = dim
        self.resample = resample
        self.correct_direction = correct_direction
        self.save_file_list = save_file_list
        if resample != None :
            self.resample.device = device
        if normalizer != None:
            self.normalizer.device = device


        for file_name in os.listdir(root_dir):            
            if 'npz' in file_name:
                self.image_name.append(file_name)
            
            elif 'pkl' in file_name:
                self.info_name.append(file_name)
                    
        self.image_name = sorted(self.image_name)
        self.info_name = sorted(self.info_name)

    def __len__(self):
        return len(self.image_name)
    
            
    def __getitem__(self, idx):
        info_file = open(os.path.join(self.root_dir,self.info_name[idx]), 'rb')
        info_data = pickle.load(info_file)
        info_file.close()
        data = np.load(os.path.join(self.root_dir, self.image_name[idx]))
        

        x = data['x']
        
        try: 
            y = data['y']
        except KeyError:
            print("y is missing")
            y = None
    
        if len(x.shape) == 3 and self.dim == '3d':
            #modality가 1개인 경우
            x = np.expand_dims(x, axis = 0)
        elif len(x.shape) == 2 and self.dim == '2d':
            x = np.expand_dims(x, axis = 0)

        if self.correct_direction:
            for i in range(len(x)):
                if i == 0:
                    result = set_direction(x[i], info_data['direction']).unsqueeze(0)
                else:
                    result = torch.vstack((result, set_direction(x[i], info_data['direction']).unsqueeze(0)))
            x = result 
            if self.mode == 'train' :
                y = set_direction(y, info_data['direction'])

        if self.normalizer:
            x = self.normalizer(x)
        
        
        if self.resample != None:
            if self.mode == 'train' :
                if isinstance(x, np.ndarray):
                    x = torch.tensor(x.copy())
                if isinstance(y, np.ndarray):
                    y = torch.tensor(y.copy())
                x, y = self.resample(x, info_data, y)
            
            
            else:
                x, _ = self.resample(x, info_data)
                
        if isinstance(x, np.ndarray):
            x = torch.tensor(x)
            

        if self.mode == 'train' :
            return x, y, info_data
        
        elif self.mode == 'test':
            return x, info_data
    
    def save_files(self, save_path, id):
        for i in range(self.__len__()):
            if self.mode == 'train':
                x, y, info_data = self.__getitem__(i)
                img_save_path = os.path.join(save_path, str(id) + '_train' + str(i) + '.npz')
                np.savez(img_save_path, x = x, y= y )
                info_save_path = os.path.join(save_path, str(id) + '_train' + str(i) + '_info.pkl')
                pickle_file = open(info_save_path, 'wb')
                pickle.dump(info_data, pickle_file)
                pickle_file.close()
            else:
                x, info_data = self.__getitem__(i)
                img_save_path = os.path.join(save_path, str(id) + '_test' + str(i) + '.npz')
                np.savez(img_save_path, x = x)
                info_save_path = os.path.join(save_path, str(id) + '_test' + str(i) + '_info.pkl')
                pickle_file = open(info_save_path, 'wb')
                pickle.dump(info_data, pickle_file)
                pickle_file.close()


class CustomDatasetV2(CustomDataset):
    def __init__(self, images_dir, label_dir = None, device = None, normalizer = None, mode = 'train', dim = '3d', correct_direction = False, resample = None,
                 patch_size = (128, 128, 128), transform = None, nifti = False):
        
        self.nifti = nifti
        self.patch_size = np.array(patch_size)
        self.transform = transform
        self.batch_size = 1
        super().__init__(device = device, normalizer = normalizer, mode = mode, dim = dim, correct_direction = correct_direction, resample = resample, root_dir = images_dir)

        if self.nifti:
            self.images_dir = images_dir
            self.label_dir = label_dir
            self.images_list = os.listdir(images_dir)
            self.labels_list = os.listdir(label_dir)

    def __len__(self):
        return super().__len__()
    
    
    def get_oversample_patch(self, images, seg):
        if not isinstance(images, np.ndarray):
            images = images.numpy()

        if not isinstance(seg, np.ndarray):
            seg = seg.numpy()

        non_zero_coords = np.array(np.where(seg != 0))
        shape_to_index = np.array(images[0].shape) - 1 
        
        oversample_high = np.clip(np.max(non_zero_coords, axis = 1) + 1, self.patch_size/2 - 1, shape_to_index - self.patch_size/2)
        oversample_low = np.clip(np.min(non_zero_coords, axis = 1), self.patch_size/2 - 1, shape_to_index - self.patch_size/2)
        oversample_center = np.random.randint(low = oversample_low, high = oversample_high + 1)

        oversample_patch_upper_bound = oversample_center + (self.patch_size/2) + 1
        oversample_patch_lower_bound = oversample_center - (self.patch_size/2 - 1)
        oversample_patch_bound = np.concatenate((oversample_patch_lower_bound.reshape(-1,1),oversample_patch_upper_bound.reshape(-1,1)), axis = 1).astype(int)

        oversample_patch_images = images[:, oversample_patch_bound[0][0] : oversample_patch_bound[0][1],
                                          oversample_patch_bound[1][0] : oversample_patch_bound[1][1],
                                          oversample_patch_bound[2][0] : oversample_patch_bound[2][1]]
        oversample_patch_seg = seg[oversample_patch_bound[0][0] : oversample_patch_bound[0][1],
                                          oversample_patch_bound[1][0] : oversample_patch_bound[1][1],
                                          oversample_patch_bound[2][0] : oversample_patch_bound[2][1]]
        
        return np.expand_dims(oversample_patch_images , axis = 0), np.expand_dims(oversample_patch_seg , axis = 0)
    
    def get_normal_patch(self, images, seg):
        if not isinstance(images, np.ndarray):
            images = images.numpy()

        if not isinstance(seg, np.ndarray):
            seg = seg.numpy()

        shape_to_index = np.array(images[0].shape) - 1 
        normal_center = np.random.randint(low = self.patch_size/2 - 1, high = shape_to_index - (self.patch_size/2) + 1 )
        normal_patch_upper_bound = normal_center + (self.patch_size/2) + 1
        normal_patch_lower_bound = normal_center - (self.patch_size/2 - 1)
        normal_patch_bound = np.concatenate((normal_patch_lower_bound.reshape(-1,1),normal_patch_upper_bound.reshape(-1,1)), axis = 1).astype(int)
        normal_patch_images = images[:, normal_patch_bound[0][0] : normal_patch_bound[0][1],
                                          normal_patch_bound[1][0] : normal_patch_bound[1][1],
                                          normal_patch_bound[2][0] : normal_patch_bound[2][1]]
        normal_patch_seg = seg[normal_patch_bound[0][0] : normal_patch_bound[0][1],
                                          normal_patch_bound[1][0] : normal_patch_bound[1][1],
                                          normal_patch_bound[2][0] : normal_patch_bound[2][1]]
        
        return np.expand_dims(normal_patch_images , axis = 0), np.expand_dims(normal_patch_seg , axis = 0)

    
    def pad_image_seg(self, images, seg): #patch size보다 전체 이미지 크기가 작을 경우 padding을 진행
        if isinstance(images, np.ndarray):
            images = torch.tensor(images.copy())
        if isinstance(seg, np.ndarray):
            seg = torch.tensor(seg.copy())
        image_shape = np.array(images[0].shape)
        if np.any(image_shape < self.patch_size):
            odd_pad = ((self.patch_size - image_shape) % 2 != 0).astype(int)
            even_pad = np.repeat(np.clip(((self.patch_size - image_shape)/2).astype(int), 0, np.inf), 2)
            for i in range(len(odd_pad)):
                even_pad[i*2] += odd_pad[i]
            even_pad = tuple(even_pad[::-1].astype(int).copy())
            images = F.pad(images, even_pad, "constant", 0 )
            seg = F.pad(seg, even_pad, "constant", 0 )

        return images, seg

    def __getitem__(self, idx):
        if self.nifti:
            info_data = dict()
            img = sitk.ReadImage(os.path.join(self.images_dir, self.images_list[idx]))
            x = sitk.GetArrayFromImage(img)
            info_data['spacing'] = img.GetSpacing()
            info_data['direction'] = img.GetDirection()

            if self.label_dir != None:
                seg = sitk.ReadImage(os.path.join(self.label_dir, self.labels_list[idx]))
                y = sitk.GetArrayFromImage(seg)
        else:
            info_file = open(os.path.join(self.root_dir,self.info_name[idx]), 'rb')
            info_data = pickle.load(info_file)
            info_file.close()
            data = np.load(os.path.join(self.root_dir, self.image_name[idx]))
            x = data['x']
            y = data['y']

        if len(x.shape) == 3 and self.dim == '3d':
            #modality가 1개인 경우
            x = np.expand_dims(x, axis = 0)
        elif len(x.shape) == 2 and self.dim == '2d':
            x = np.expand_dims(x, axis = 0)

        if self.correct_direction:
            for i in range(len(x)):
                if i == 0:
                    result = set_direction(x[i], info_data['direction']).unsqueeze(0)
                else:
                    result = torch.vstack((result, set_direction(x[i], info_data['direction']).unsqueeze(0)))
            x = result 
            if self.mode == 'train' :
                y = set_direction(y, info_data['direction'])

        if self.normalizer:
            x = self.normalizer(x)
        
        
        if self.resample != None:
            if self.mode == 'train' :
                if isinstance(x, np.ndarray):
                    x = torch.tensor(x.copy())
                if isinstance(y, np.ndarray):
                    y = torch.tensor(y.copy())
                x, y = self.resample(x, info_data, y)
            
            else:
                x, _ = self.resample(x, info_data)
        
        x, y = self.pad_image_seg(x, y)
        if self.batch_size == 1 :
            result_images, result_seg = self.get_normal_patch(x, y)


        for i_b in range(self.batch_size//2):
            oversample_images, oversample_seg = self.get_oversample_patch(x, y)
            normal_images, normal_seg = self.get_normal_patch(x, y)

            if i_b == 0:
                result_images = np.concatenate((normal_images, oversample_images), axis = 0) #bs, n, x, y, z
                result_seg = np.concatenate((normal_seg, oversample_seg), axis = 0) #bs, x, y, z
            
            else:
                result_images = np.concatenate((result_images, oversample_images), axis = 0)
                result_seg = np.concatenate((result_seg, oversample_seg), axis = 0)
                result_images = np.concatenate((result_images, normal_images), axis = 0)
                result_seg = np.concatenate((result_seg, normal_seg), axis = 0)


        if isinstance(result_images, np.ndarray):
            result_images = torch.tensor(result_images)
        
        if isinstance(result_seg, np.ndarray):
            result_seg = torch.tensor(result_seg)
        
        if result_images.device.index != self.device:
            result_images = result_images.to(self.device)

        if result_seg.device.index != self.device:
            result_seg = result_seg.to(self.device)

        if self.transform:
            result_images, result_seg = self.transform(result_images.to(self.device), result_seg.to(self.device), info_data)
        
        return result_images.squeeze(0), result_seg.squeeze(0).long(), info_data



class Dataset2point5(CustomDataset):
    def __init__(self, root_dir, degree, device=None, normalizer=None, mode='train', dim='3d', correct_direction=True, resample=None):
        super().__init__(root_dir, device=device, normalizer=normalizer, mode=mode, dim=dim, correct_direction=correct_direction, resample=resample)
        self.transform = Batch_Affine_3D([degree], [0], is_random = False)
    
    def set_direction(self, img, direction):
        return super().set_direction(img, direction)
    
    def __len__(self):
        return super().__len__()

    def __getitem__(self, idx):
        images, seg, info_data = super().__getitem__(idx) # images => n_modalities, x, y, z, seg도 마찬가지
        
        if self.transform != None:
            if isinstance(images, np.ndarray):
                images = np.expand_dims(images, aixs = 0)
            else:
                images = images.unsqueeze(0)

            if isinstance(seg, np.ndarray):
                seg = np.expand_dims(seg, axis = 0)
            else:
                seg = seg.unsqueeze(0)

            images = self.transform(images)
            seg = self.transform(seg)
            while len(images.shape) > 4:
                images = images.squeeze(0)
            while len(seg.shape) > 3:
                seg = seg.squeeze(0)


        return images, seg
    



class DataLoader3D(object):
    def __init__(self, dataset, num_iteration = 2, patch_size = (480,256,256), transform= None, device = 0, batch_size = 2, seg_one_hot = False, is_half = False, info_class = None):
        self.dataset = dataset
        self.current_index = 0
        self.patch_size = np.array(patch_size)
        self.batch_size = batch_size
        self.now_iter = 0
        self.images, self.seg, self.info = self.get_image_seg(self.current_index)
        self.device = device
        self.pass_patient_index = []
        self.num_iteration = num_iteration
        self.seg_one_hot = seg_one_hot
        self.is_half = is_half
        self.info_class = info_class #[0, 1, 3, 4] 이런식
        self.transform = transform

    
    def new_epoch(self):
        self.now_iter = 0
        self.current_index = 0
    
    def is_end(self):
        return self.current_index == len(self.dataset) and self.now_iter == 0
    
    def return_class(self):
        return np.unique(self.seg)

    def next_index(self):
        if self.now_iter == self.num_iteration:
            self.images, self.seg, self.info = self.get_image_seg(self.current_index)
            self.current_index += 1
            self.now_iter = 0
        
        else:
            self.now_iter += 1

    def get_image_seg(self, current_index): #patch size보다 전체 이미지 크기가 작을 경우 padding을 진행
        images, seg, info = self.dataset[current_index]
        if isinstance(images, np.ndarray):
            images = torch.tensor(images.copy())
        if isinstance(seg, np.ndarray):
            seg = torch.tensor(seg.copy())
        image_shape = np.array(images[0].shape)[-3:]
        if np.any(image_shape < self.patch_size):
            odd_pad = (np.max((np.array([0,0,0]),(self.patch_size - image_shape)), axis = 0) % 2 != 0).astype(int)
            even_pad = np.repeat(np.clip(((self.patch_size - image_shape)/2).astype(int), 0, np.inf), 2)
            for i in range(len(odd_pad)):
                even_pad[i*2] += odd_pad[i]
            even_pad = tuple(even_pad[::-1].astype(int).copy())
            images = F.pad(images, even_pad, "constant", 0 )
            seg = F.pad(seg, even_pad, "constant", 0 )

        return images, seg, info


    def get_oversample_patch(self, non_zero_coords, shape_to_index):
        oversample_high = np.clip(np.max(non_zero_coords, axis = 1) + 1, self.patch_size/2 - 1, shape_to_index - self.patch_size/2).reshape(1,3) + 1
        oversample_low = np.clip(np.min(non_zero_coords, axis = 1), self.patch_size/2 - 1, shape_to_index - self.patch_size/2).reshape(1,3)
        oversample_high = oversample_high.astype(int)
        oversample_low = oversample_low.astype(int)
        oversample_center = np.random.randint(oversample_low, oversample_high)

        oversample_patch_upper_bound = oversample_center + (self.patch_size/2) + 1
        oversample_patch_lower_bound = oversample_center - (self.patch_size/2 - 1)
        oversample_patch_bound = np.concatenate((oversample_patch_lower_bound.reshape(-1,1),oversample_patch_upper_bound.reshape(-1,1)), axis = 1).astype(int)
        oversample_patch_images = self.images[:, oversample_patch_bound[0][0] : oversample_patch_bound[0][1],
                                          oversample_patch_bound[1][0] : oversample_patch_bound[1][1],
                                          oversample_patch_bound[2][0] : oversample_patch_bound[2][1]]
        oversample_patch_seg = self.seg[oversample_patch_bound[0][0] : oversample_patch_bound[0][1],
                                          oversample_patch_bound[1][0] : oversample_patch_bound[1][1],
                                          oversample_patch_bound[2][0] : oversample_patch_bound[2][1]]
        
        return np.expand_dims(oversample_patch_images , axis = 0), np.expand_dims(oversample_patch_seg , axis = 0)

    def get_normal_patch(self, non_zero_coords, shape_to_index):
        normal_center = np.random.randint(low = self.patch_size/2 - 1, high = shape_to_index - (self.patch_size/2) + 1 )
        normal_patch_upper_bound = normal_center + (self.patch_size/2) + 1
        normal_patch_lower_bound = normal_center - (self.patch_size/2 - 1)
        normal_patch_bound = np.concatenate((normal_patch_lower_bound.reshape(-1,1),normal_patch_upper_bound.reshape(-1,1)), axis = 1).astype(int)
        normal_patch_images = self.images[:, normal_patch_bound[0][0] : normal_patch_bound[0][1],
                                          normal_patch_bound[1][0] : normal_patch_bound[1][1],
                                          normal_patch_bound[2][0] : normal_patch_bound[2][1]]
        normal_patch_seg = self.seg[normal_patch_bound[0][0] : normal_patch_bound[0][1],
                                          normal_patch_bound[1][0] : normal_patch_bound[1][1],
                                          normal_patch_bound[2][0] : normal_patch_bound[2][1]]
        
        return np.expand_dims(normal_patch_images , axis = 0), np.expand_dims(normal_patch_seg , axis = 0)
    



    def generate_train_batch(self):
        non_zero_coords = np.array(np.where(self.seg != 0))
        shape_to_index = np.array(self.images[0].shape) - 1 

        if self.batch_size == 1 :
            result_images, result_seg = self.get_normal_patch(non_zero_coords, shape_to_index)


        for i_b in range(self.batch_size//2): # oversample batch를 2/3 으로 바꿔야함.
            oversample_images, oversample_seg = self.get_oversample_patch(non_zero_coords, shape_to_index)
            normal_images, normal_seg = self.get_normal_patch(non_zero_coords, shape_to_index)

            if i_b == 0:
                result_images = np.concatenate((normal_images, oversample_images), axis = 0) #bs, n, x, y, z
                result_seg = np.concatenate((normal_seg, oversample_seg), axis = 0) #bs, x, y, z
            
            else:
                result_images = np.concatenate((result_images, oversample_images), axis = 0)
                result_seg = np.concatenate((result_seg, oversample_seg), axis = 0)
                result_images = np.concatenate((result_images, normal_images), axis = 0)
                result_seg = np.concatenate((result_seg, normal_seg), axis = 0)

        self.next_index()
        
    
        
        result_images = torch.Tensor(result_images).to(self.device)
        result_seg = torch.Tensor(result_seg).to(self.device)


        if self.transform:
            result_images, result_seg = self.transform(result_images, result_seg, self.info)
        
        if self.seg_one_hot:
            result_seg = F.one_hot(result_seg.long(),  num_classes = len(self.info_class)).permute(0, 4, 1, 2, 3)

        if self.is_half:
            result_images = result_images.half()
    

        return result_images, result_seg.long()


"""
class DataLoader2D(object):
    def __init__(self, dataset, batch_size = 16, full_volume_size = (600, 600, 720), transform = None):
        self.current_index = 0
        self.images, self.seg, self.info = 
        self.transform = transform
        self.current_batch = 0
        self.batch_size = batch_size
    
    def div_batch(self):
        # batch를 배수에 맞게 나눈다.
        # mat_batch를 정한다.
        
    
    def next_index(self):
        if self.current_batch == self.n_batch:
            self.images, self.seg, self.info = self.get_image_seg(self.current_index)
            self.current_index += 1
            self.now_iter = 0
        
        else:
            self.now_iter += 1

    def get_image_seg(self, current_index): #patch size보다 전체 이미지 크기가 작을 경우 padding을 진행
        images, seg, info = self.dataset[current_index]
        if isinstance(images, np.ndarray):
            images = torch.tensor(images.copy())
        if isinstance(seg, np.ndarray):
            seg = torch.tensor(seg.copy())
        image_shape = np.array(images[0].shape)[-3:]
        if np.any(image_shape < self.patch_size):
            odd_pad = (np.max((np.array([0,0,0]),(self.patch_size - image_shape)), axis = 0) % 2 != 0).astype(int)
            even_pad = np.repeat(np.clip(((self.patch_size - image_shape)/2).astype(int), 0, np.inf), 2)
            for i in range(len(odd_pad)):
                even_pad[i*2] += odd_pad[i]
            even_pad = tuple(even_pad[::-1].astype(int).copy())
            images = F.pad(images, even_pad, "constant", 0 )
            seg = F.pad(seg, even_pad, "constant", 0 )
        if len(images.shape)

        return images, seg, info
    
    def sampling_by_axis(self):
        # 축을 따라 sampling

        return
    

    def new_epoch(self):
        self.now_iter = 0
        self.current_index =

    def padding(self, image, seg):
        # 전체 Image volume을 가져올 때 padding을 진행한다.

    def get_image_seg(self, current_index):
        return super().get_image_seg(current_index) 0
    
    def next_index(self):
        if self.now_iter == self.num_iteration:
            self.images, self.seg, self.info = self.get_image_seg(self.current_index)
            self.current_index += 1
            self.now_iter = 0
        
        else:
            self.now_iter += 1

    def cal_batch(self):
        

    def generate_train_batch(self):


        
        if self.transform:
            result_images, result_seg = self.transform(result_images, result_seg, self.info)

"""



class Interactive_DataLoader(DataLoader3D):
    def __init__(self, dataset, num_iteration = 1, patch_size = (480,256,256), batch_size = 2, device = None, transform = None):
        self.dataset = dataset
        self.current_index = 0
        self.patch_size = np.array(patch_size)
        self.images, self.seg, self.info = self.get_image_seg(self.current_index)
        self.device = device
        self.num_iteration = num_iteration
        self.now_iter = 0
        self.batch_size = batch_size
        self.transform = transform
        
    def get_image_seg(self, current_index):
        return super().get_image_seg(current_index)
    
    def next_index(self):
        return super().next_index()
    
    def is_end(self):
        return super().is_end()
    
    def new_epoch(self):
        return super().new_epoch()
    
    def get_oversample_patch(self, non_zero_coords, shape_to_index):
        return super().get_oversample_patch(non_zero_coords, shape_to_index)


    """
    def generate_train_batch(self):
        oversample_images = np.expand_dims(self.images, axis = 0)
        oversample_seg = np.expand_dims(self.seg, axis = 0)
        oversample_images = torch.Tensor(oversample_images).to(self.device)
        oversample_seg = torch.Tensor(oversample_seg).to(self.device)

        if self.transform != None: 
            oversample_images, oversample_seg = self.transform(oversample_images, oversample_seg, self.info)
        oversample_seg = self.sampling(oversample_seg)

        while len(oversample_images.shape) < len(['bs', 'n_modalities', 'D', 'H', 'W']) :
            oversample_images = oversample_images.unsqueeze(0)
        self.next_index()

        return oversample_images, oversample_seg

    def sampling(self, seg):
        category = torch.unique(seg)
        random_index = torch.randint(low = 1, high = len(category), size =(1,)).to(seg.device.index)
        new_seg = (seg == category[random_index]).long()
        
        return new_seg

    """
    def generate_train_batch(self):
        non_zero_coords = np.array(np.where(self.seg != 0))
        shape_to_index = np.array(self.images[0].shape) - 1 
        oversample_images, oversample_seg = self.get_oversample_patch(non_zero_coords, shape_to_index) #bs, n, x, y,z // bs, x, y, z
        oversample_images = torch.Tensor(oversample_images).to(self.device)
        oversample_seg = torch.Tensor(oversample_seg).to(self.device)
        if self.transform != None: 
            oversample_images, oversample_seg = self.transform(oversample_images, oversample_seg, self.info)
            
        oversample_images, oversample_seg = self.sampling(oversample_images.squeeze(0), oversample_seg)
        self.next_index()

        return oversample_images, oversample_seg
        

        if self.transform != None: 
            oversample_images, oversample_seg = self.transform(oversample_images, oversample_seg, self.info)
            
        oversample_images, oversample_seg = self.sampling(oversample_images.squeeze(0), oversample_seg)
        self.next_index()

        return oversample_images, oversample_seg
        
    
    def sampling(self, images, seg):
        
        category = torch.unique(seg)
        for i in range(self.batch_size):
            random_index = torch.randint(low = 1, high = len(category), size =(1,)).to(seg.device.index)
            if i == 0:
                new_seg = (seg == category[random_index]).long()

            else:
                new_seg = torch.vstack((new_seg, (seg == category[random_index]).long()))
        new_images = torch.tile(images, [self.batch_size, *(torch.tensor(images.shape)//torch.tensor(images.shape))])

        return new_images, new_seg
    




    

"""
class DataLoader2D(object):
    def __init__(self, patch_size = , batch_size = 8, ):


    def new_epoch(self):
        self.now_iter = 0
        self.current_index = 0
    
    def is_end(self):
        return self.current_index == len(self.dataset) and self.now_iter == 0
    
    def return_class(self):
        return np.unique(self.seg)

    def next_index(self):
        if self.now_iter == self.num_iteration:
            self.images, self.seg, self.info = self.get_image_seg(self.current_index)
            self.current_index += 1
            self.now_iter = 0
        
        else:
            self.now_iter += 1

    def generate_train_batch(self):
"""




from ingradient_library.preprocessing import Cropping
import SimpleITK as sitk
import pickle
import os
import numpy as np
from ingradient_library.preprocessing import Resampling
import torch

class Data_Organizer(object):
    # Resampling 후 저장 추가해야 됨.
    def __init__(self, SAVE_PATH, ID = 'ingradient', resampling = None, normalizer = None, mode = 'train', do_set_direction = False):
        self.crop = Cropping()
        self.SAVE_PATH = SAVE_PATH
        self.ID = ID
        self.resampling = resampling
        self.normalizer = normalizer
        self.mode = mode
        self.do_set_direction = do_set_direction
    
    def set_direction(self, img, direction):
        direction = np.array(direction).reshape(3,3)
        inverse = np.where(direction == -1)[0]
        permute = np.where(direction != 0)[1]
        if 0 in inverse:
            img = img[::-1, :, :]

        if 1 in inverse:
            img = img[:, ::-1, :]

        if 2 in inverse:
            img = img[:, :, ::-1]

    def run(self, seg_path, img_path_list, index, modality = ['CT']):
        seg = sitk.ReadImage(seg_path)
        save_dict = dict()
        save_dict['spacing'] = seg.GetSpacing()
        save_dict['direction'] = seg.GetDirection()
        save_dict['origin'] = seg.GetOrigin()
        save_dict['modality'] = modality
        if 'CT' in modality:
            self.crop.is_CT = True
        else:
            self.crop.is_CT = False
        seg = sitk.GetArrayFromImage(seg)
        save_path = os.path.join(self.SAVE_PATH, self.ID+str(index)+'_info.pkl')
        pickle_file = open(save_path, 'wb')
        pickle.dump(save_dict, pickle_file)
        pickle_file.close()

        for i in range(len(img_path_list)):
            img_path = img_path_list[i]
            img = sitk.ReadImage(img_path)
            img = sitk.GetArrayFromImage(img)
            if len(img.shape) != 4:
                img = np.expand_dims(img, axis = 0)
            if i == 0:
                images = img
            else:
                images = np.vstack(images, img)
        
        
        print("BEFORE Cropping, current path : ", seg_path, "image shapes :", images.shape, "seg shapes :", seg.shape, "spacing :", save_dict['spacing'])
        img_arr, seg_arr = self.crop(images, seg)

        if self.do_set_direction:
            for i in range(len(images)):
                images[i] = self.set_direction(images[i], save_dict['direction'])
            
                if self.mode == 'train' :
                    seg = self.set_direction(seg, save_dict['direction'])
        
        if self.normalizer != None:
            img_arr = self.normalizer(img_arr)

        

        if self.resampling != None:
            if isinstance(img_arr, np.ndarray):
                img_arr = torch.tensor(img_arr.copy())
            else:
                img_arr = img_arr
            
            if isinstance(seg_arr, np.ndarray):
                seg_arr = torch.tensor(seg_arr.copy())
            
            else:
                seg_arr = seg_arr

            if self.mode == 'train' :
                result = self.resampling(torch.vstack((img_arr, seg_arr.unsqueeze(0))), save_dict)
                img_arr = result[:-1]
                seg_arr = result[-1].squeeze(0)
            
            else:
                img_arr = self.resampling(img_arr, save_dict)
        

        print("current path : ", seg_path, "image shapes :", img_arr.shape, "seg shapes :", seg_arr.shape, "spacing :", save_dict['spacing'])
        if isinstance(img_arr, np.ndarray):
            np.savez(os.path.join(self.SAVE_PATH, self.ID+str(index)+'.npz'), x =img_arr, y=seg_arr)
        
        else:
            np.savez(os.path.join(self.SAVE_PATH, self.ID+str(index)+'.npz'), x =img_arr.numpy(), y=seg_arr.numpy())

    


import cv2
import os
import cc3d
import SimpleITK as sitk
import numpy as np
import time

class Manual_Labeling(object):
    """
    왼쪽 화면 : predicted mask
    오른쪽 화면 : grount truth
    p : 이전 파일
    n : 다음 파일
    좌클릭 : 선택할 component
    우클릭 : 삭제할 component
    r : 현재 선택 된 component
    s : 결과 저장
    """
    def __init__(self, root_dir, save_dir, gt_path = None):
        self.root_dir = root_dir
        self.save_dir = save_dir
        self.gt_path = gt_path
        self.file_list = sorted(os.listdir(root_dir))
        if gt_path != None:
            self.gt_list = sorted([f for f in os.listdir(gt_path) if 'npz' in f ])
            self.sorted_by_gt()
    
    def sorted_by_gt(self):
        temp = []
        for g in self.gt_list:
            for f in self.file_list:
                prefix_g = g.split('.')[0]
                prefix_f = f.split('s')[0]
            
                if prefix_g == prefix_f:
                    temp.append(f)
                    break
        
        self.file_list = temp
        
    def get_coordinates(self, x, y):
        if self.selected_axis == 0:
            return [self.axis_index, y, x]
        
        if self.selected_axis == 1:
            return [y, self.axis_index, x]
        
        if self.selected_axis == 2:
            return [y, x, self.axis_index]
        
    

    def mouse_event(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDBLCLK:
            self.update_result(self.get_coordinates(x, y))
        
        elif event == cv2.EVENT_RBUTTONDBLCLK:
            self.update_result(self.get_coordinates(x, y), mode = 'right')
    
        elif event == cv2.EVENT_MOUSEWHEEL:
            if flags > 0 and self.axis_index < self.image.shape[self.selected_axis]:
                self.axis_index += 1
                self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
            elif self.axis_index > 0 and flags < 0:
                self.axis_index -= 1
                self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
        
    def save(self, image, index):
        print("Start saving the file. file name : ", self.file_list[index])
        nii_img = sitk.GetImageFromArray(image)
        sitk.WriteImage(nii_img, os.path.join(self.save_dir, self.file_list[index]))
        print("File saving is complete. file name : ", self.file_list[index])
    
    def update_result(self, point, mode = 'left'):
        if mode == 'left':
            self.result += (self.cc_mask == self.cc_mask[point[0]][point[1]][point[2]])
        
        else:
            self.result[np.where(self.cc_mask == self.cc_mask[point[0]][point[1]][point[2]])] = 0
    
    def run(self, name = 'left = pred || right = gt'):
        self.name = name
        cv2.namedWindow(name)
        cv2.setMouseCallback(name, self.mouse_event)
        index = 0
        self.selected_axis = 0
        self.load_nifti_img(index)
        self.axis_index = self.image.shape[self.selected_axis] // 2

        while isinstance(self.image, np.ndarray):
            self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
            key = cv2.waitKey()
            if key == ord('n'):
                print('next image')
                index += 1
                self.load_nifti_img(index)
                self.axis_index = self.image.shape[self.selected_axis] // 2
                self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
            
            if key == ord('p'):
                if index > 0:
                    print('prev image')
                    index -= 1
                    self.load_nifti_img(index)
                    self.axis_index = self.image.shape[self.selected_axis] // 2
                    self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
            
            if key == ord('v'):
                self.selected_axis += 1
                self.selected_axis %= 3
                self.axis_index = self.image.shape[self.selected_axis] // 2
                self.imshow_by_axis(self.image, self.axis_index, self.selected_axis, self.name)
            
            if key == ord('r'):
                self.imshow_by_axis(self.result, self.axis_index, self.selected_axis, self.name)
                wait_key = cv2.waitKey(3000)#pauses for 3 seconds before fetching next image
                if wait_key == 27:#if ESC is pressed, exit loop
                    cv2.destroyAllWindows()
                    break
                elif wait_key == ord('s'):
                    self.save(self.result, index)
                    time.sleep(1)
            if key == ord('s'):
                self.save(self.result, index)
                time.sleep(1)
                


    def imshow_by_axis(self, image, axis_index, selected_axis, name):
        if selected_axis == 0:
            img_to_show = image[axis_index]
            gt = self.gt[axis_index]
        if selected_axis == 1:
            img_to_show = image[:, axis_index]
            gt = self.gt[:, axis_index]
        if selected_axis == 2:
            img_to_show = image[:, :, axis_index]
            gt = self.gt[:, :, axis_index]
        mean_val = img_to_show.mean()
        std_val = img_to_show.std()
        img_to_show = (img_to_show - mean_val) / std_val
        img_to_show = np.hstack((img_to_show, gt))
        cv2.imshow(name, img_to_show)


    def load_nifti_img(self, index):
        if index >= len(self.file_list) :
            return False
        print('predicted file name : ', self.file_list[index])
        sitk_img = sitk.ReadImage(os.path.join(self.root_dir, self.file_list[index]))
        img_arr = sitk.GetArrayFromImage(sitk_img)
        mask = (img_arr != 0)
        conected_component_mask = cc3d.connected_components(mask, connectivity = 26)
        self.result = np.zeros(conected_component_mask.shape)
        
        self.image = img_arr
        self.cc_mask = conected_component_mask
        
        if self.gt_path != None:
            print('gt file name : ', self.gt_list[index])
            self.gt = np.load(os.path.join(self.gt_path, self.gt_list[index]))
            self.gt = self.gt['y']


from ingradient_library.patch_transform import *


def get_transform_params(device):
    affine_kwargs = {'degree':[30,30,30], 'axis':[0,1,2], 'scale':[0.7, 1.4], 'use_gpu': True, 'device':device}
    noise_kwargs = {'device': device, 'prob_per_modalities': 0.5}
    blur_kwargs = {'sigma': 1.4, 'width':3, 'device':device}
    contrast_kwargs = {'contrast_range' : [0.65, 1.5], 'preserve_range' : True, 'device':device}
    gamma_kwargs = {'gamma_range': (0.5, 1.5), 'epsilon':1e-7, 'device':device, 'retain_stats':True}
    bright_kwargs = {'device':device , 'rng': [0.7, 1.3]}
    mirror_kwargs = {'x_prob':0.5, 'y_prob':0.5,'z_prob':0.5}
    affine_prob = 0.2
    noise_prob = 0.15
    blur_prob = 0.2
    contrast_prob = 0.15
    gamma_prob = 0.15
    bright_prob = 0.15
    mirror_prob = 0.5
    lowres_prob = 0.2


    affine_transform = Batch_Affine_3D(**affine_kwargs)
    noise_transform = Batch_Gaussian_Noise(**noise_kwargs)
    blur_transform = Batch_Gaussian_Blur_3D(**blur_kwargs)
    contrast_transform = Batch_Contrast(**contrast_kwargs)
    bright_transform = Batch_Brightness(**bright_kwargs)
    gamma_transform = Batch_GammaTransform(**gamma_kwargs)
    mirror_transform = Batch_Mirroring(**mirror_kwargs)
    lowres_transform = Batch_Low_Resolution()

    return [affine_transform, noise_transform, blur_transform, contrast_transform, bright_transform, gamma_transform, mirror_transform, lowres_transform], [affine_prob, noise_prob, blur_prob, contrast_prob, gamma_prob, bright_prob, mirror_prob, lowres_prob]


def get_interactive_params(device):
    affine_kwargs = {'degree':[10,10,10], 'axis':[0,1,2], 'scale':[0.75, 1.25], 'use_gpu': True, 'device':device}
    mirror_kwargs = {'x_prob':0.5, 'y_prob':0.5,'z_prob':0.5}

    affine_prob = 0.66
    mirro_prob = 0.50
    affine_transform = Batch_Affine_3D(**affine_kwargs)
    mirror_transform = Batch_Mirroring(**mirror_kwargs)

    return [affine_transform, mirror_transform], [affine_prob, mirro_prob]


def get_maic_params():
    device = None
    affine_kwargs = {'degree':[10,10,10], 'axis':[0,1,2], 'scale':[0.98, 1.02], 'use_gpu': True, 'device':device}
    noise_kwargs = {'device': device, 'prob_per_modalities': 0.5}
    blur_kwargs = {'sigma': 1.4, 'width':3, 'device':device}
    contrast_kwargs = {'contrast_range' : [0.65, 1.5], 'preserve_range' : True, 'device':device}
    gamma_kwargs = {'gamma_range': (0.5, 1.5), 'epsilon':1e-7, 'device':device, 'retain_stats':True}
    bright_kwargs = {'device':device , 'rng': [0.7, 1.3]}
    mirror_kwargs = {'x_prob':0.5, 'y_prob':0.5,'z_prob':0.5}
    affine_prob = 0.45
    noise_prob = 0.15
    blur_prob = 0.2
    contrast_prob = 0.15
    gamma_prob = 0.15
    bright_prob = 0.15
    mirror_prob = 0.5
    lowres_prob = 0.2


    affine_transform = Batch_Affine_3D(**affine_kwargs)
    noise_transform = Batch_Gaussian_Noise(**noise_kwargs)
    blur_transform = Batch_Gaussian_Blur_3D(**blur_kwargs)
    contrast_transform = Batch_Contrast(**contrast_kwargs)
    bright_transform = Batch_Brightness(**bright_kwargs)
    gamma_transform = Batch_GammaTransform(**gamma_kwargs)
    mirror_transform = Batch_Mirroring(**mirror_kwargs)
    lowres_transform = Batch_Low_Resolution()

    return [affine_transform, noise_transform, blur_transform, contrast_transform, bright_transform, gamma_transform, mirror_transform, lowres_transform], [affine_prob, noise_prob, blur_prob, contrast_prob, gamma_prob, bright_prob, mirror_prob, lowres_prob]


import os
from ingradient_library.data_organizer import Data_Organizer

def medical_decathlon_organizer(SAVE_PATH, LOAD_PATH, id = 'Task'):
    TR_LOAD_PATH = os.path.join(LOAD_PATH, 'imagesTr')
    SEG_LOAD_PATH = os.path.join(LOAD_PATH, 'labelsTr')

    img_file = [e for e in os.listdir(TR_LOAD_PATH) if e[0] != '.']
    seg_file = [e for e in os.listdir(SEG_LOAD_PATH) if e[0] != '.']

    img_file = sorted(img_file)
    seg_file = sorted(seg_file)    
    organizer = Data_Organizer(SAVE_PATH, id)

    for i in range(len(img_file)):
        seg_file_path = os.path.join(SEG_LOAD_PATH, seg_file[i])
        img_file_path = os.path.join(TR_LOAD_PATH, img_file[i])
        organizer.run(seg_file_path, [img_file_path], i)





from torch.optim.lr_scheduler import _LRScheduler

class PolynomialLRDecay(_LRScheduler):
    """Polynomial learning rate decay until step reach to max_decay_step
    
    Args:
        optimizer (Optimizer): Wrapped optimizer.
        max_decay_steps: after this step, we stop decreasing learning rate
        end_learning_rate: scheduler stoping learning rate decay, value of learning rate must be this value
        power: The power of the polynomial.
    """
    
    def __init__(self, optimizer, max_decay_steps, end_learning_rate=1e-5, power=2.0):
        if max_decay_steps <= 1.:
            raise ValueError('max_decay_steps should be greater than 1.')
        self.max_decay_steps = max_decay_steps
        self.end_learning_rate = end_learning_rate
        self.power = power
        self.last_step = 0
        super().__init__(optimizer)
        
    def get_lr(self):
        if self.last_step > self.max_decay_steps:
            return [self.end_learning_rate for _ in self.base_lrs]

        return [(base_lr - self.end_learning_rate) * 
                ((1 - self.last_step / self.max_decay_steps) ** (self.power)) + 
                self.end_learning_rate for base_lr in self.base_lrs]
    
    def step(self, step=None):
        if step is None:
            step = self.last_step + 1
        self.last_step = step if step != 0 else 1
        if self.last_step <= self.max_decay_steps:
            decay_lrs = [(base_lr - self.end_learning_rate) * 
                         ((1 - self.last_step / self.max_decay_steps) ** (self.power)) + 
                         self.end_learning_rate for base_lr in self.base_lrs]
            for param_group, lr in zip(self.optimizer.param_groups, decay_lrs):
                param_group['lr'] = lr


import torch
import torch.nn.functional as F
import torch.nn as nn

class Deep_Supervision_Loss(nn.Module):
    """
    1. 모든 Loss Funtion은 Reduction이어야한다.
    2. Target (Batch Size, x, y, z) 의 형태이며 element는 class number이다.
    3. Input은 (Batch Size, n_class, x, y, z) 형태이다.

    """
    def __init__(self, n_deepsupervision, loss1, loss2 = None, regularizer = 1.0, deep_supervision_weight = 0.5):
        super(Deep_Supervision_Loss, self).__init__()
        self.n_deepsupervision = n_deepsupervision
        loss1.reduction = 'none'
        if loss2 != None:
            loss1.reduction = 'none'
        self.loss1 = loss1
        self.loss2 = loss2
        self.regularizer = regularizer
        self.deep_supervision_weight = deep_supervision_weight
        self.writer = SummaryWriter()
        
    
    def forward(self, output, target):
        batch_size, deep_supervision_size, channel_size, size_x, size_y, size_y = output.shape
        output = output.reshape(batch_size * deep_supervision_size, channel_size, -1)
        target = target.reshape(batch_size, 1, -1).repeat(1, deep_supervision_size,1).reshape(batch_size * deep_supervision_size, -1)

        result = self.loss1(output, target)
        if isinstance(self.loss1, nn.CrossEntropyLoss):
            result = result.mean(-1)
        self.writer.add_scalar('Iteration_Loss/loss1', result.mean())
        if self.loss2 != None:
            loss2 = self.loss2(output, target)
            if isinstance(self.loss2, nn.CrossEntropyLoss):
                loss2 = loss2.mean(-1)
                self.writer.add_scalar('Iteration_Loss/loss2', loss2.mean())
            result += loss2 * self.regularizer

        result = result.view(batch_size, deep_supervision_size, -1)
    
        weight = torch.ones(deep_supervision_size).to(output.device.index)
        acc_value = self.deep_supervision_weight
        for i in range(deep_supervision_size):
            weight[i] = acc_value ** i
    
        weight = weight / weight.sum()
        weight = weight.view(1,deep_supervision_size ,1)
        result = result * weight

        return result.mean()
        


class Binary_Dice_Loss(nn.Module):
    def __init__(self, smooth = 1.0):
        super(Binary_Dice_Loss, self).__init__()
        self.smooth = smooth
    
    def forward(self, output, target):
        # output shape => (n_bs, 1, -1)
        # target shape => (n_bs, -1)
        n_bs = output.shape[0]
        output = output.view(n_bs, -1)
        target = target.view(n_bs, -1)
        intersection = (output * target).sum(-1)  # (n_bs, n_ds)
        union = output.sum(-1) + target.sum(-1)
        dice = 2.0 * (intersection + self.smooth) / (union + self.smooth)
        return 1 - dice


class Multi_Dice_Loss(nn.Module):
    def __init__(self, num_classes, smooth = 0.0, epsilon = 1e-8, background_index = 0):
        super(Multi_Dice_Loss, self).__init__()
        self.smooth = smooth
        self.background_index = background_index
        self.num_classes = num_classes
        self.mask = (torch.arange(0, num_classes) != self.background_index)
        self.reduction = 'none'
        self.epsilon = epsilon

    def forward(self, output, target):
        #output shape => (n_bs, n_classes, -1)
        #target shape => (n_bs, -1)
        self.mask.to(output.device.index)
        n_bs = output.shape[0]
        output = output.view(n_bs, self.num_classes, -1)
        target = target.view(n_bs, -1)
        output = output[:, self.mask].permute(0, 2, 1) # (n_bs,  -1, n_cls)
        target = F.one_hot(target, num_classes = self.num_classes)
        target = target[:,:, self.mask]
        intersection = (output * target).sum((-1,-2)) #각 class 별 intersection score를 합함.
        union = output.sum((-1,-2)) + target.sum((-1,-2)) #각 class 별 union score를 합함.

        dice = 2.0 * (intersection + self.smooth) / (union + self.smooth + self.epsilon)
        
        return 1 - dice





import revlib
import torch.nn as nn
import torch

# UNet의 경우 병렬적인 구조가 너무 많음. DeepLabV3 의 3D 모델 구현으로 가는 방안이 나아보임.
# Active Contour Loss를 Reversible 하게 설계하는 것도 좋은 방법. 현재는 마지막 Volume에 *3을 하게 됨.
# 해당 아이디어는 다음과 같은데 결국 Active Contour Loss는 F를 Sobel_Filter_x(X), Sobel_FIlter_y(X), Soble_Filter_z(X) 하는거고
# 이 결과를 각각 Sx, Sy, Sz라고 하면 Sx**2 + Sy**2 + Sz**2 를 구해서 Length로 사용하는데, 위의 Sobel Filter를 각기 구하는 과정에서
# 이를 G3(G2(G1(X))) = Sx**2 + Sy**2 + Sz**2로 수렴하게 하면 됨. => random 변수로 형성 후, Neural Network 학습 이후 뽑아내면 됨.
# 장점 No patch Needed 그냥 한번에 돌리면 됨.

def conv_block(in_channels, out_channels, kernel_size = 3, padding = 1, stride = 1, dilation = 1):
    return nn.Sequential([nn.Conv3d(in_channels, out_channels, kernel_size = kernel_size, padding = padding, stride = stride, dilation = dilation),
                        nn.InstanceNorm3d(out_channels),
                        nn.Dropout3d(0.2),
                        nn.LeakyReLU()])

def block(in_channels, kernel_size = 3, padding = 1, stride = 1, dilation = 1):
    return nn.Sequential([conv_block(in_channels, in_channels*2, kernel_size = kernel_size, padding = padding, stride = stride, dilation = dilation),
                        conv_block(in_channels*2, in_channels, kernel_size = 3, padding = 1, stride = 1)])

def classifier(in_channels, out_channels):
    return nn.Sequential(conv_block(in_channels, in_channels),
                         nn.Conv3d(in_channels, out_channels, 1, 1, 0))

def backbone(in_channels, depth = 5, variation = 6):
    return nn.Sequential(nn.Sequential([block(in_channels, dilation = 1, padding = 1)] +
            [block(in_channels, dilation= i*variation, padding = i*variation) for i in range(depth, 0, -1)]))

def model(in_channels, out_channels, depth, variation):
    return revlib.ReversibleSequential([backbone(in_channels, depth, variation), classifier(in_channels, out_channels)])


import os
from ingradient_library.nnunet_3D_run import *
from ingradient_library.inference import Inference

LOAD_PATH = '/home/ubuntu/ingradient_lib_test'
SAVE_MODEL_PATH = '/home/ubuntu/model_state_dict'
patch_size = (40, 56, 40)
batch_size = 8
n_classes = 3
n_modalities = 1
GPU_DEVICE_INDEX = 0
SAVE_RESULT_PATH ='/home/ubuntu/result_file'
MODEL_LOAD = os.path.join(SAVE_MODEL_PATH,'epoch50_model_state_dict.pkl' )

narrow_inference_run(LOAD_PATH, MODEL_LOAD, SAVE_RESULT_PATH, GPU_DEVICE_INDEX, patch_size, n_classes, n_modalities)
#nnunet_3D_run_narrow(LOAD_PATH, SAVE_MODEL_PATH, GPU_DEVICE_INDEX, patch_size, batch_size, n_classes, n_modalities)


from torch.utils.tensorboard import SummaryWriter
import torch.optim as optim
from torch.utils.data import random_split
import torch
import matplotlib.pyplot as plt
import time
from ingradient_library.dataloads import *
from ingradient_library.model import *
from ingradient_library.preprocessing import *
from ingradient_library.deep_supervision_loss import *
from ingradient_library.visualization import *
from ingradient_library.optimizer import SAMSGD
from ingradient_library.sampling import *




"""

class TrainerV2(obejct):
    def __init__(self, tr_dataloader, model, optimizer, losses, scheduler, activiation = torch.softmax, n_deepsupervision_layer = 4, val_dataloader = None, n_epoch = 1000, save_path = None):
        self.tr_dataloader = tr_dataloader
        self.model = model
        self.optimizer = optimizer
        self.losses = losses
        self.scheduler = scheduler
        self.n_deepsupervision_layer = n_deepsupervision_layer
        self.val_dataloader = val_dataloader
        self.n_epoch = n_epoch
        self.save_path = save_path
        self.activation = 
    
    def run(self):
        writer = SummaryWriter()
        current_iter = 0
        val_iter = 0
        for e in range(self.n_epoch):
            self.tr_dl.new_epoch()
            n_iter = 0
            self.model.train()
            epoch_tr_loss = 0
            epoch_tr_div = 0
            epoch_val_loss = 0
            epoch_val_div = 0
            while not self.tr_dl.is_end():
                current_iter += 1
                n_iter += 1
                images, seg = self.tr_dl.generate_train_batch()
                if isinstance(self.optimizer, SAMSGD):
                    def closure():
                        self.optimizer.zero_grad()
                        output = self.model(images)
                        for ds in self.n_deepsupervision_layer:
                            output[ds] 
                        loss = self.calculate_loss(output, seg, writer, current_iter)
                        loss.backward()
                        return loss
                    self.optimizer.step(closure)
                    epoch_tr_loss = 1
                    epoch_tr_div = 1
                
                else:
                    output = self.model(images)
                    self.optimizer.zero_grad()
                    loss = self.calculate_loss(output, seg, writer, current_iter)
                    loss.backward()
                    self.optimizer.step()
                    epoch_tr_loss += loss.item()
                    epoch_tr_div += 1

            if self.val_dl != None:
                self.val_dl.new_epoch()
                val_loss = 0
                n_iter = 0
                self.model.eval()
                while not self.val_dl.is_end():
                    val_iter +=1
                    n_iter +=1
                    with torch.no_grad():
                        images, seg = self.val_dl.generate_train_batch()
                        output = self.model(images)
                        loss = self.calculate_loss(output, seg, writer, val_iter, True)
                        epoch_val_loss += loss.item()
                        epoch_val_div += 1

            writer.add_scalar('epoch\Loss_train', epoch_tr_loss/epoch_tr_div , e) #(batch, -1) 로 나오게끔 resize
            if self.val_dl != None:
                writer.add_scalar('epoch\Loss_val', epoch_val_loss/epoch_val_div , e) #(batch, -1) 로 나오게끔 resize

            if self.scheduler != None: 
                self.scheduler.step()
            if self.save_path != None:
                file_name = 'epoch'+ str(e) + '_model_state_dict.pkl'
                torch.save(self.model.state_dict(), os.path.join(self.save_path, file_name))
        
"""


class Trainer(object):
    def __init__(self, tr_dataloader, model,  optimizer, losses, regulizers = [], is_deep_supervision = False,
                deep_supervision_weight = 0.5, n_epoch = 1000, val_dataloader = None, save_path = None, scheduler = None, visualization = False):
 
        self.model = model
        self.visualization = visualization
        self.optimizer = optimizer
        self.n_epoch = n_epoch
        self.save_path = save_path
        self.visualization = visualization
        self.tr_dl = tr_dataloader
        self.val_dl = val_dataloader
        self.scheduler = scheduler
        self.losses = losses
        self.is_deep_supervision = is_deep_supervision
        self.deep_supervision_weight = deep_supervision_weight
        self.regulizers  = regulizers
        if len(self.regulizers) == 0:
            self.regulizers = [1] * len(self.losses)
            
    def calculate_loss(self, output, target, writer, epoch, is_val = False):
        if self.is_deep_supervision == True:
            batch_size = output.shape[0]
            deep_supervision_size = output.shape[1]
            channel_size = output.shape[2]
            output = output.reshape(batch_size * deep_supervision_size, channel_size, -1).contiguous()
            if channel_size == 1:
                target = target.reshape(batch_size, 1, -1).repeat(1, deep_supervision_size,1).reshape(batch_size * deep_supervision_size, 1, -1).contiguous()
                target = target.float()
            else:
                target = target.reshape(batch_size, 1, -1).repeat(1, deep_supervision_size,1).reshape(batch_size * deep_supervision_size, -1).contiguous()
            weight = torch.ones(deep_supervision_size).to(output.device.index)
            weight.requires_grad = False
            acc_value = self.deep_supervision_weight
            for i in range(deep_supervision_size):
                weight[i] = acc_value ** i
            
            weight = weight / weight.sum()
            weight = weight.view(1,deep_supervision_size ,1).contiguous()
        
        else:
            batch_size = output.shape[0]
            channel_size = output.shape[1]
            output = output.reshape(batch_size, channel_size, -1).contiguous()
            if channel_size == 1:
                target = target.float()
                target = target.reshape(batch_size, 1, -1).contiguous()
            
            else:
                target = target.reshape(batch_size, -1).contiguous()

        for i in range(len(self.losses)):
            loss = self.losses[i] #여러 Loss들에 대해
            temp = loss(output, target).contiguous() * torch.tensor(self.regulizers[i]).float().to(output.device.index)
            if i == 0:
                losses = temp
            else:
                losses += temp

            #for _ in range(len(temp.shape)-1):
            #    temp = temp.mean(-1)
           
            if is_val:
                writer.add_scalar('Iteration\Val_Loss_'+ str(type(loss)), temp.mean().item(), epoch) #(batch, -1) 로 나오게끔 resize
            else:
                writer.add_scalar('Iteration\Train_Loss_'+ str(type(loss)), temp.mean().item(), epoch) #(batch, -1) 로 나오게끔 resize
                    
        if self.is_deep_supervision == True:
            result = (losses * weight).mean()
            
        else:
            result = losses.mean()

        return result.contiguous()
    


    def load_model_state_dict(self, path, load_classifier = False):
        pretrained_dict = torch.load(path)
        if not load_classifier:
            classifier = list(self.model._modules)[-1]
            for k in list(pretrained_dict.keys()):
                if classifier in k:
                    del pretrained_dict[k]

        current_dict = self.backbone.state_dict()
        current_dict.update(pretrained_dict)
        self.backbone.load_state_dict(current_dict)

    def run(self):
        writer = SummaryWriter()
        current_iter = 0
        val_iter = 0
        for e in range(self.n_epoch):
            self.tr_dl.new_epoch()
            n_iter = 0
            self.model.train()
            epoch_tr_loss = 0
            epoch_tr_div = 0
            epoch_val_loss = 0
            epoch_val_div = 0
            while not self.tr_dl.is_end():
                current_iter += 1
                n_iter += 1
                images, seg = self.tr_dl.generate_train_batch()
                if isinstance(self.optimizer, SAMSGD):
                    def closure():
                        self.optimizer.zero_grad()
                        output = self.model(images)
                        loss = self.calculate_loss(output, seg, writer, current_iter)
                        loss.backward()
                        return loss
                    self.optimizer.step(closure)
                    epoch_tr_loss = 1
                    epoch_tr_div = 1
                
                else:
                    output = self.model(images)
                    self.optimizer.zero_grad()
                    loss = self.calculate_loss(output, seg, writer, current_iter)
                    loss.backward()
                    self.optimizer.step()
                    epoch_tr_loss += loss.item()
                    epoch_tr_div += 1

                if self.tr_dl.current_index < 3 and not isinstance(self.optimizer, SAMSGD) and self.visualization:
                    print("###########", self.tr_dl.current_index, "###########")
                    n_classes = 1
                    if not self.is_deep_supervision:
                        print('output')
                        n_classes = output.shape[2]
                        plt.imshow(torch.argmax(output[0,:,:,:,output.shape[-1]//2], dim = 0).detach().cpu(), vmax =  n_classes, vmin= 0)
                        plt.show()
                    elif not isinstance(self.optimizer, SAMSGD):
                        print('output')
                        n_classes = output.shape[1]
                        plt.imshow(torch.argmax(output[0,0,:,:,:,output.shape[-1]//2], dim = 0).detach().cpu(), vmax = n_classes, vmin= 0)
                        plt.show()
                    print('images')
                    plt.imshow(images[0,0,:,:,output.shape[-1]//2].detach().cpu())
                    plt.show()
                    print('ground_truth')
                    plt.imshow(seg[0,:,:,output.shape[-1]//2].detach().cpu(), vmax = n_classes , vmin= 0)
                    plt.show()


            if self.val_dl != None:
                self.val_dl.new_epoch()
                val_loss = 0
                n_iter = 0
                self.model.eval()
                while not self.val_dl.is_end():
                    val_iter +=1
                    n_iter +=1
                    with torch.no_grad():
                        images, seg = self.val_dl.generate_train_batch()
                        output = self.model(images)
                        loss = self.calculate_loss(output, seg, writer, val_iter, True)
                        epoch_val_loss += loss.item()
                        epoch_val_div += 1

            writer.add_scalar('epoch\Loss_train', epoch_tr_loss/epoch_tr_div , e) #(batch, -1) 로 나오게끔 resize
            if self.val_dl != None:
                writer.add_scalar('epoch\Loss_val', epoch_val_loss/epoch_val_div , e) #(batch, -1) 로 나오게끔 resize

            if self.scheduler != None: 
                self.scheduler.step()
            if self.save_path != None:
                file_name = 'epoch'+ str(e) + '_model_state_dict.pkl'
                torch.save(self.model.state_dict(), os.path.join(self.save_path, file_name))
        


class Interactive_Trainer(Trainer):
    def __init__(self, tr_dataloader, model, optimizer, scheduler, losses, regulizers=[], is_deep_supervision=False, deep_supervision_weight=0.5, n_epoch=1000, val_dataloader=None, save_path=None, n_point = 10 ):
        super().__init__(tr_dataloader, model, optimizer, scheduler, losses, regulizers=regulizers, is_deep_supervision=is_deep_supervision, deep_supervision_weight=deep_supervision_weight, n_epoch=n_epoch, val_dataloader=val_dataloader, save_path=save_path)
        self.n_point = 10

    def calculate_loss(self, output, target, writer, epoch):
        return super().calculate_loss(output, target, writer, epoch)
    
    def run(self, visualization_ineterval = 5, summary_writer_path = None):
        if summary_writer_path != None:
            writer = SummaryWriter(summary_writer_path)
        else:
            writer = SummaryWriter()
        for e in range(self.n_epoch):
            self.tr_dl.new_epoch()
            train_loss = 0
            n_iter = 0
            self.model.train()
            
            while not self.tr_dl.is_end():
                n_iter += 1
                images, seg = self.tr_dl.generate_train_batch()
                bs, _, nx, ny, nz = images.shape #bs, n_modalities, 
                #First
                positive_points, negative_points = point_sampler(seg, first = True)
                output = self.model(images, positive_points, negative_points)
                self.optimizer.zero_grad()
                loss = self.calculate_loss(output, seg, writer, e)
                train_loss = loss.item()
                
                center = output.shape[3] // 2
                
                loss.backward()
                self.optimizer.step()

                for i in range(self.n_point):
                    zero_grad = time.time()
                    self.optimizer.zero_grad()
                    print('zero grad', time.time() - zero_grad)
                    prev_seg_time = time.time()
                    prev_seg = output.detach()
                    print('prev_seg', time.time()- prev_seg_time)
                    ps_time = time.time()
                    point_sampler(seg, prev_seg, positive_points, negative_points)
                    out_time = time.time()
                    print('ps time', time.time()- ps_time)
                    output = self.model(images, positive_points, negative_points, prev_seg)
                    loss_time = time.time()
                    print('out time', time.time()- out_time)
                    loss = self.calculate_loss(output, seg, writer, e)
                    train_loss = loss.item()
                    if self.tr_dl.current_index % visualization_ineterval == 0:
                        print("points : ",i + 1)
                        visualization(output, seg)
                    loss.backward()
                    self.optimizer.step()
                    print('out time', time.time()- loss_time)
            self.scheduler.step()


            if self.save_path != None:
                file_name = 'epoch'+ str(e) + '_model_state_dict.pkl'
                torch.save(self.model.state_dict(), os.path.join(self.save_path, file_name))



from ingradient_library.patch_transform import *
from ingradient_library.preprocessing import Resampling

class Transform(object):
    def __init__(self, transforms, probs):
        self.probs = probs
        self.transforms = transforms
        self.type_img_transformer = [Batch_Gaussian_Noise, Batch_Gaussian_Blur_3D, Batch_Contrast, Batch_Brightness, Batch_GammaTransform, Batch_Low_Resolution]
        self.type_img_seg_transformer = [Batch_Affine_3D, Batch_Mirroring, Batch_Random_Permute]
        self.type_resampling = [Resampling]
        self.img_transformer = []
        self.img_seg_transformer = []
        self.resampling = []
        self.img_transformer_prob = []
        self.img_seg_transformer_prob = []

        for transform, prob in zip(self.transforms, probs):
            if type(transform) in self.type_img_transformer:
                self.img_transformer.append(transform)
                self.img_transformer_prob.append(prob)
            elif type(transform) in self.type_img_seg_transformer:
                self.img_seg_transformer.append(transform)
                self.img_seg_transformer_prob.append(prob)
            elif type(transform) in self.type_resampling:
                if prob != 1:
                    raise ValueError ("The probablity of Resampling msut be 1.0")
                self.resampling.append(transform)
    

    def __call__(self, images, seg, info):
        for i in range(len(self.resampling)):
            images = self.resampling[i](images, info, mode = 'x')
            seg = self.resampling[i](seg, info, mode = 'y')

        for i in range(len(self.img_transformer)):
            current_prob = self.img_transformer_prob[i]
            if np.random.uniform(0, 1) < current_prob:
                images = self.img_transformer[i](images)
    
        for i in range(len(self.img_seg_transformer)):
            current_prob = self.img_seg_transformer_prob[i]
            if np.random.uniform(0, 1) < current_prob:
                if isinstance(self.img_seg_transformer[i], Batch_Affine_3D):
                    self.img_seg_transformer[i].get_matrices_and_coords(images)
                
                else:
                    self.img_seg_transformer[i].get_random_variables()
                images = self.img_seg_transformer[i](images)
                seg = self.img_seg_transformer[i](seg)


        return images, seg


import torch
import numpy as np
import pickle
import os
import SimpleITK as sitk
import torch.nn.functional as F
import cc3d
import copy



class CT_Lung_Cropping(object):
    def __init__(self, clip = False, threshold_clip = [-900, -200 ], threshold_cut = [-1000, -350]):
        self.clip = clip
        self.threshold_clip = threshold_clip
        self.threshold_cut = threshold_cut
    
    def __call__(self, image):
        if self.clip:
            clip_img = np.clip(image, self.threshold_clip[0], self.threshold_clip[1])
        else:
            clip_img = image
        mask =  (clip_img > self.threshold_cut[0]) * (clip_img < self.threshold_cut[1])
        if not isinstance(mask, np.ndarray):
            mask = mask.numpy()
        conected_component_mask = cc3d.connected_components(mask, connectivity = 26)
        count = np.unique(conected_component_mask, return_counts = True)[1]
        lung_index = np.argsort(count)[::-1]
        if len(lung_index) > 2:
            lung_index = lung_index[2]
        
        else:
            lung_index = lung_index[-1]
        lung_max = np.max(np.where(conected_component_mask == lung_index), axis = 1)
        lung_min = np.min(np.where(conected_component_mask == lung_index), axis = 1)
        
        return lung_min, lung_max

class Normalizer(object):
    def __init__(self, percentile_clipping_ratio = None):
        self.percentile = percentile_clipping_ratio
        self.device = None

    def __call__(self, images):
        if isinstance(images, np.ndarray):
            images = torch.Tensor(images)
        
        if images.device.index != self.device and self.device:
            images = images.to(self.device)
        
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
        
        images = images.float()
        if self.percentile != None:
            images = self.percentile_clipping(images)
        n_modalities, nx, ny, nz = images.shape
        mmean = torch.mean(images.reshape(n_modalities, -1), -1, True).reshape(n_modalities,1,1,1)
        mstd = torch.std(images.reshape(n_modalities, -1), -1, True).reshape(n_modalities,1,1,1)
        
        return (images - mmean) / mstd
        
    
    def percentile_clipping(self,images):
        n_modalities, nx, ny, nz = images.shape
        images = images.reshape(n_modalities, -1)
        vals, indices = torch.sort(images, -1)
        upper_index = indices[:, int(indices.shape[1] * self.percentile[1]):]
        lower_index = indices[:, :int(indices.shape[1] * self.percentile[0])]
        upper_val = vals[:, int(indices.shape[1] * self.percentile[1])]
        lower_val = vals[:, int(indices.shape[1] * self.percentile[0])]

        for i in range(n_modalities):
            images[i, upper_index[i]] = upper_val[i]
            images[i, lower_index[i]] = lower_val[i]
        
        return images.reshape(n_modalities, nx, ny, nz)


class Min_Max_Normalizer(object):
    def __init__(self, min = -1024, max = 3024, device = None):
        self.min = min
        self.max = max
        self.device = device
    def __call__(self, images):
        if isinstance(images, np.ndarray):
            images = torch.Tensor(images)
        
        if images.device.index != self.device and self.device:
            images = images.to(self.device)
        
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
        
        images = 2 * ((images - self.min) / (self.max - self.min)) - 1
        return images

class Fixed_Normalizer(object):
    def __init__(self, mean = -155.58, std = 324.70, min = -1024, max = 3025, device = None):
        self.mean = mean
        self.std = std
        self.min = min
        self.max = max
        self.device = device
    
    def __call__(self, images):
        if isinstance(images, np.ndarray):
            images = torch.Tensor(images)
        
        if images.device.index != self.device and self.device:
            images = images.to(self.device)
        
        if len(images.shape) == 3:
            images = images.unsqueeze(0)
        
        images = images.float()
        images = torch.clip(images, self.min, self.max)
        images = images - self.mean
        images = images / self.std
        return images

class Cropping(object):
    def __call__(self, images, seg = None, mode = 'train', is_CT = True):
        temp = copy.deepcopy(images)
        if is_CT:
            temp[np.where(temp < - 700)] = 0
        non_zero_index = np.where(temp.astype(int) != 0)
        min_val = np.min(non_zero_index, axis = 1)
        max_val = np.max(non_zero_index, axis = 1)
        if mode == 'train':
            return images[:, min_val[-3]:max_val[-3]+1, min_val[-2]:max_val[-2]+1, min_val[-1]:max_val[-1]+1], seg[ min_val[-3]:max_val[-3]+1, min_val[-2]:max_val[-2]+1, min_val[-1]:max_val[-1]+1]
        if mode == 'test':
            return images[:, min_val[-3]:max_val[-3]+1, min_val[-2]:max_val[-2]+1, min_val[-1]:max_val[-1]+1]


class Resampling(object):
    def __init__(self, anisotropy_axis_index = None, target_spacing = [1.0, 1.0, 1.0], random = False, device = None, spacing_order = [-3, -2, -1]):
        self.device = device
        self.anisotropy_axis_index = self.new_anisotropy_axis(anisotropy_axis_index)
        self.target_spacing = np.array(target_spacing)
        self.random = random
        self.spacing_order = spacing_order
    
    def new_anisotropy_axis(self, array):
        for i in range(len(array)):
            if array[i] == 2:
                array[i] = 0
            else:
                array[i] += 1
        return array


    def __call__(self, data, info_data, mode = 'x'):
        if isinstance(data, np.ndarray):
            data = torch.from_numpy(data.copy())
        if data.device.index != self.device and self.device:
            data = data.to(self.device)

        original_spacing = info_data['spacing']
        if self.random:
            self.target_spacing = original_spacing * np.random.uniform(0.20, 5.0, size = np.array(original_spacing).shape)
        new_shape = list((original_spacing/self.target_spacing * np.array([data.shape[self.spacing_order[0]], data.shape[self.spacing_order[1]], data.shape[self.spacing_order[2]]])).astype(int))
        data = F.interpolate(data.unsqueeze(0), new_shape).squeeze(0)

        return data



class Get_target_spacing(object):
    def __init__(self, anisotropy_threshold = 3.0, image_dimension = 3):
        self.spaces = None
        self.first = True
        self.anisotropy_threshold = anisotropy_threshold
        self.isotropy_percentile_value = 50
        self.anisotropy_percentile_value = 90
        self.image_dimension = image_dimension

    def run(self, dir_path, dtype = 'pkl'):
        """
        type은 pkl 또는 nifti
        """
        for file_name in os.listdir(dir_path):
            if dtype == 'pkl':
                if file_name[-3:] == 'pkl':
                    path = os.path.join(dir_path, file_name)
                    file = open(path, 'rb')
                    self.append(pickle.load(file)['spacing'])
                    file.close()
            
            if dtype == 'nifti':
                if 'nii' in file_name:
                    path = os.path.join(dir_path, file_name)
                    img = sitk.ReadImage(path)
                    spacing = img.GetSpacing()
                    self.append(spacing)
    
    def run_file_path(self, file_path):
        if type == 'pkl':
                if file_path[-3:] == 'pkl':
                    file = open(file_path, 'rb')
                    self.append(pickle.load(file)['spacing'])
                    file.close()
        if type == 'nifti':
            if 'nii' in file_path:
                img = sitk.ReadImage(file_path)
                spacing = img.GetSpacing()
                self.append(spacing)
    
    def reset(self):
        self.spaces = None

    def append(self, item):
        item = np.array(item).reshape(1,self.image_dimension)
        if self.first:
            self.spaces = item
            self.first = False
        else:
            self.spaces = np.vstack((self.spaces, item)) 
    
    def get_target_space(self):
        anisotropy_axis_index = self.is_anisotropy()
        isotropy_axis_index = np.arange(self.image_dimension) != anisotropy_axis_index
        iso = np.percentile(self.spaces, self.isotropy_percentile_value, 0)
        aniso = np.percentile(self.spaces, self.anisotropy_percentile_value, 0)
        result = np.zeros(self.image_dimension)
        result[anisotropy_axis_index] = aniso[anisotropy_axis_index]
        result[isotropy_axis_index] = iso[isotropy_axis_index]
        return result, anisotropy_axis_index
    
    def is_anisotropy(self):
        return np.where((np.max(self.spaces, axis = 0) / np.min(self.spaces, axis = 0)) > self.anisotropy_threshold)[0].astype(int)
    


class ResamplingV2(object):
    def __init__(self, anisotropy_axis_index = [], target_spacing = [1.0, 1.0, 1.0], random = False, debug_shape = False,
                 device = None, spacing_order = [-3, -2, -1], do_check = False, check_threshold = 120):
        self.device = device
        self.anisotropy_axis_index = self.new_anisotropy_axis(anisotropy_axis_index)
        self.target_spacing = np.array(target_spacing)
        self.random = random
        self.spacing_order = spacing_order
        self.do_check = do_check
        self.check_threshold = check_threshold
        self.debug_shape = debug_shape
    
    def new_anisotropy_axis(self, array):
        for i in range(len(array)):
            if array[i] == 2:
                array[i] = 0
            else:
                array[i] += 1
        return array

    def correct_dims(self, shape):
        shape = np.array(shape)
        if np.any(shape) < self.check_threshold:
            shape[np.where(shape < self.check_threshold)] = np.max(shape) // 2
        return tuple(shape)

    def __call__(self, img, info_data, seg = None):
        if isinstance(img, np.ndarray):
            img = torch.from_numpy(img.copy())
        if img.device.index != self.device and self.device:
            img = img.to(self.device)
        
        if isinstance(seg, np.ndarray):
            seg = torch.from_numpy(seg.copy())
        if img.device.index != self.device and self.device:
            seg = img.to(self.device)
        
        if isinstance(info_data, dict):
            original_spacing = info_data['spacing']
        else:
            original_spacing = info_data
        
        if self.random:
            self.target_spacing = original_spacing * np.random.uniform(0.20, 5.0, size = np.array(original_spacing).shape)
        new_shape = list((original_spacing/self.target_spacing * np.array([img.shape[self.spacing_order[0]], img.shape[self.spacing_order[1]], img.shape[self.spacing_order[2]]])).astype(int))

        if self.do_check:
            new_shape = self.correct_dims(new_shape)
        
        img = F.interpolate(img.unsqueeze(0), new_shape, mode = 'trilinear').squeeze(0)
        
        if seg != None:
            seg = F.interpolate(seg.unsqueeze(0).unsqueeze(0), new_shape, mode = 'nearest').squeeze(0).squeeze(0)

        if self.debug_shape:
            print("img shape : ", img.shape, "seg shape : ", seg.shape)
            img_shape = np.array(img.shape[-3:])
            seg_shape = np.array(seg.shape[-3:])
            
            if np.any(img_shape != seg_shape):
                print("Waring, Resampling was wrong")
            
        return img, seg




import torch
import torch.nn.functional as F
import torch.nn as nn

class Binary_Dice_Loss(nn.Module):
    def __init__(self, smooth = 1.0):
        super(Binary_Dice_Loss, self).__init__()
        self.smooth = smooth
    
    def forward(self, output, target):
        # output shape => (n_bs, 1, -1)
        # target shape => (n_bs, -1)
        n_bs = output.shape[0]
        output = output.view(n_bs, -1)
        target = target.view(n_bs, -1)
        intersection = (output * target).sum(-1)  # (n_bs, n_ds)
        union = output.sum(-1) + target.sum(-1)
        dice = 2.0 * (intersection + self.smooth) / (union + self.smooth)
        dice = dice.mean()
        return 1 - dice


class Multi_Dice_Loss(nn.Module):
    def __init__(self, num_classes, smooth = 0.0, epsilon = 1e-8, background_index = 0):
        super(Multi_Dice_Loss, self).__init__()
        self.smooth = smooth
        self.background_index = background_index
        self.num_classes = num_classes
        self.mask = (torch.arange(0, num_classes) != self.background_index)
        self.reduction = 'none'
        self.epsilon = epsilon

    def forward(self, output, target):
        #output shape => (n_bs, n_classes, -1)
        #target shape => (n_bs, -1)
        self.mask.to(output.device.index)
        n_bs = output.shape[0]
        output = output.view(n_bs, self.num_classes, -1)
        target = target.view(n_bs, -1)
        output = output[:, self.mask].permute(0, 2, 1) # (n_bs,  -1, n_cls)
        target = F.one_hot(target, num_classes = self.num_classes)
        target = target[:,:, self.mask]
        intersection = (output * target).sum((-1,-2)) #각 class 별 intersection score를 합함.
        union = output.sum((-1,-2)) + target.sum((-1,-2)) #각 class 별 union score를 합함.

        dice = 2.0 * (intersection + self.smooth) / (union + self.smooth + self.epsilon)
        dice = dice.mean()
        return 1 - dice




import torch.optim as optim
from torch.utils.data import random_split
import torch
from dataloads import *
from unet3d import *
from preprocessing import *
from data_augmentation import *
from deepsupervision_loss import *
from visualization import *

DATASET_PATH = '/content/gdrive/MyDrive/DeepVibe/Datasets/BRATS2020'
SAVED_MODEL = '/content/gdrive/MyDrive/DeepVibe/Model State Dictionary/3D UNet with Deep Supervision Weight Loss/epoch_14_deepsupervision_UNet3d.pkl'
HISTORY_PATH = '/content/gdrive/MyDrive/DeepVibe/Model State Dictionary/3D UNet with Deep Supervision Weight Loss'

torch.manual_seed(0)
ds = CustomDataset(DATASET_PATH, normalizer=Normalizer())
tr_ds, val_ds = random_split(ds, [len(ds) - len(ds)//9, len(ds)//9])
da = DataAugmentation()
tr_dl = DataLoader3D(tr_ds, da, batch_size = 2, num_iteration = 1)
val_dl = DataLoader3D(val_ds, batch_size = 2, num_iteration = 1)

model = UNet3DDeepsupervision().to(0)
model_dict = torch.load('/content/gdrive/MyDrive/DeepVibe/Model State Dictionary/3D UNet with Deep Supervision Weight Loss/epoch_21_deepsupervision_UNet3d.pkl')
model.load_state_dict(model_dict)
optimizer = optim.Adam(params = model.parameters(), lr = 0.01)
epoch = 1000

with open(os.path.join(HISTORY_PATH,"epoch_history.pkl"), "rb") as fp:   #Pickling
    train_loss, valid_loss = pickle.load(fp)

for e in range(15, epoch):
    optimizer.param_groups[0]['lr'] = 0.01 * ((e/epoch)**0.9)
    tr_dl.new_epoch()
    iter_loss = []
    model.train()
    while not tr_dl.is_end():
        optimizer.zero_grad()
        images, seg = tr_dl.generate_train_batch()
        output = model(images)
        loss = deep_supervision_loss(output, seg)
        iter_loss.append(loss.item())
        loss.backward()
        optimizer.step()
    
    val_dl.new_epoch()
    val_iter_loss = []
    model.eval()
    visualization_index = 24
    while not val_dl.is_end():
        with torch.no_grad():
            images, seg = val_dl.generate_train_batch()
            output = model(images)
            loss = deep_supervision_loss(output, seg)
            val_iter_loss.append(loss.item())
            if val_dl.current_index == visualization_index:
                print("iteration : ", val_dl.current_index + 1)
                deep_supervision_visualization(output, seg)

    
    file_name = 'epoch_'+ str(e) + '_deepsupervision_UNet3d.pkl'
    torch.save(model.state_dict(), os.path.join(HISTORY_PATH,file_name))
    train_loss.append(np.array(iter_loss).mean())
    valid_loss.append(np.array(val_iter_loss).mean())
    print("epoch :", e+1, " Train Loss :",train_loss[-1], "Valid Loss :", valid_loss[-1])
    with open(os.path.join(HISTORY_PATH, "epoch_history.pkl"), "wb") as fp:
        pickle.dump([train_loss, valid_loss], fp)
        



import torch
import numpy as np
from ingradient_library.patch_transform import *

class nnUNet_DataAugmentation(object):
    def __init__(self, affine_kwargs = None, noise_kwargs = None, blur_kwargs = None, contrast_kwargs = None,
                 gamma_kwargs = None, bright_kwargs = None, mirror_kwargs = None, device = 0,
                 affine_prob = 0.2, noise_prob = 0.15, blur_prob = 0.2, contrast_prob = 0.15, gamma_prob = 0.15, bright_prob = 0.15, mirror_prob = 0.5):

#    def __init__(self, gamma_range = (0.5, 2), epsilon = 1e-7, device = None, retain_stats = True):
        if affine_kwargs == None:
            affine_kwargs = {'degree':[30,30,30], 'axis':[0,1,2], 'scale':[0.85, 1.15], 'use_gpu': True, 'device':device}
        
        if noise_kwargs ==  None:
            noise_kwargs = {'device': device, 'prob_per_modalities': 0.5}
        
        if blur_kwargs == None:
            blur_kwargs = {'sigma': 1.4, 'width':3, 'device':device}
        
        if contrast_kwargs == None:
            contrast_kwargs = {'contrast_range' : [0.65, 1.5], 'preserve_range' : True, 'device':device}
        
        if gamma_kwargs == None:
            gamma_kwargs = {'gamma_range': (0.5, 1.5), 'epsilon':1e-7, 'device':device, 'retain_stats':True}
        
        if bright_kwargs == None:
            bright_kwargs = {'device':device , 'rng': [0.7, 1.3]}
        
        if mirror_kwargs == None:
            mirror_kwargs = {'x_prob':0.5, 'y_prob':0.5,'z_prob':0.5}

        self.device= device
        self.affine_transform = Batch_Affine_3D(**affine_kwargs)
        self.noise_transform = Batch_Gaussian_Noise(**noise_kwargs)
        self.blur_transform = Batch_Gaussian_Blur_3D(**blur_kwargs)
        self.contrast_transform = Batch_Contrast(**contrast_kwargs)
        self.bright_transform = Batch_Brightness(**bright_kwargs)
        self.gamma_transform = Batch_GammaTransform(**gamma_kwargs)
        self.mirror_transform = Batch_Mirroring(**mirror_kwargs)
        
        self.affine_prob = affine_prob
        self.noise_prob = noise_prob
        self.blur_prob = blur_prob
        self.contrast_prob = contrast_prob
        self.bright_prob = bright_prob
        self.gamma_prob = gamma_prob
        self.mirror_prob = mirror_prob


    def __call__(self, images, seg):
        images = images.float()
        seg = seg.float()
        if images.device.index != self.device:
            images = images.to(self.device)

        if seg.device.index != self.device:
            seg = seg.to(self.device)
        

        affine_prob = np.random.uniform(0, 1)
        noise_prob = np.random.uniform(0, 1)
        blur_prob = np.random.uniform(0, 1)
        contrast_prob = np.random.uniform(0, 1)
        bright_prob = np.random.uniform(0, 1)
        gamma_prob = np.random.uniform(0, 1)
        mirror_prob = np.random.uniform(0,1)

        
        if affine_prob < self.affine_prob:
            self.affine_transform.get_matrices_and_coords(images)
            images = self.affine_transform(images)
            seg = self.affine_transform(seg)
        
        if noise_prob < self.noise_prob:
            images = self.noise_transform(images)
        
        if blur_prob < self.blur_prob:
            images = self.blur_transform(images)

        if contrast_prob < self.contrast_prob:
            images = self.contrast_transform(images)

        if bright_prob < self.bright_prob:
            images = self.bright_transform(images)
        
        if gamma_prob < self.gamma_prob:
            images = self.gamma_transform(images)
        
        if mirror_prob < self.mirror_prob:
            self.mirror_transform.get_mirror_axis()
            images = self.mirror_transform(images)
            seg = self.mirror_transform(seg)
        
        

        return images, seg

        


        

        


        


import numpy as np
import torch
import time

def is_overlap(points, current_point):
    if points[current_point[0], current_point[1], current_point[2], current_point[3]] != 0:
        return True
    else:
        return False

"""
def get_random_sample(points, candidates, batch_index):
    n_candidates = len(candidates[0])
    index = np.random.choice(np.arange(n_candidates), size = (n_candidates,), replace = False)
    p_batch = batch_index
    i = 0
    x = candidates[-3][index[i]]
    y = candidates[-2][index[i]]
    z = candidates[-1][index[i]]
    
    while is_overlap(points, current_point=[p_batch, x, y, z]) and i < len(index) - 1:
        i += 1
        p_batch = batch_index
        x = candidates[-3][index[i]]
        y = candidates[-2][index[i]]
        z = candidates[-1][index[i]]
    
    point = [p_batch, x, y, z]
    if not is_overlap(points, point):
        points[point[0], point[1], point[2], point[3]] = 1
    return points
"""

"""

def get_random_sample(points, candidates, batch_index):
    n_candidates = len(candidates[0])
    index = np.random.choice(np.arange(n_candidates), size = (1,), replace = False)

    x = candidates[-3][index[0]]
    y = candidates[-2][index[0]]
    z = candidates[-1][index[0]]
    
    point = [batch_index, x, y, z]
    points[point[0], point[1], point[2], point[3]] = 1
    return points
"""
def point_sampler(target, output = None, positive_points = None, negative_points = None, first = False):
    # target shpae, prev_seg => (batch, x, y, z)
    # positive_points, negative_points => (n, 4) [(batch_index, x_index, y_index, z_index), ...과 같은 방식], 입력 시 Transpose를 이용해 쉽게 넣을 수 있다.
    # 초기값 세팅
    bs, x, y, z = target.shape
    
    id_time = time.time()
    if type(output) == type(None):
        prev_seg = torch.zeros(target.shape).to(target.device.index)
    
    else:
        prev_seg = (output[:, 0] > 0.5).long()
        
    if type(positive_points) == type(None):
        positive_points = torch.zeros(target.shape).to(target.device.index)
    
    if type(negative_points) == type(None):
        negative_points = torch.zeros(target.shape).to(target.device.index)

    print("id time", time.time()- id_time)
    
    for batch_index in range(bs):
        if (prev_seg[batch_index] != target[batch_index]).sum() == 0:
            # Ground truth = Predicted Segmentation
            continue

        elif prev_seg[batch_index].sum() >= target[batch_index].sum():
            # negative sampling
            cand_time = time.time()
            candidates = torch.where((target * (prev_seg != target)) == 0)
            n_candidates = len(candidates[0])
            index = torch.randint(low = 0, high = n_candidates, size = (1,))
            negative_points[batch_index, candidates[-3][index], candidates[-2][index],candidates[-1][index]] = 1
            print("find cand", time.time()- cand_time)

        elif prev_seg[batch_index].sum() < target[batch_index].sum():
            # positive sampling
            cand_time = time.time()
            candidates = torch.where((target * (prev_seg != target)) == 1)
            n_candidates = len(candidates[0])
            index = torch.randint(low = 0, high = n_candidates, size = (1,))
            positive_points[batch_index, candidates[-3][index], candidates[-2][index],candidates[-1][index]] = 1
            print("find cand", time.time()- cand_time)

    if first:
        return positive_points, negative_points


def get_filter(arr_size, r):
    center = np.array(arr_size)//2
    coords = np.ogrid[:arr_size[0], :arr_size[1], :arr_size[2]]
    distance = np.sqrt((coords[0] - center[0])**2 + (coords[1]-center[1])**2 + (coords[2]-center[2])**2) 
    return torch.tensor((distance <= r*1.1)).unsqueeze(0).unsqueeze(0).long()





import numpy as np
import torch
import torch.nn as nn


class ACLoss(nn.Module):
    """
    Active Contour Loss
    based on sobel filter
    """

    def __init__(self, miu=1.0, classes=3):
        super(ACLoss, self).__init__()

        self.miu = miu
        self.classes = classes
        sobel_x = np.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]])
        sobel_y = np.array([[-1, -2, -1], [0, 0, 0], [1, 2, 1]])

        self.sobel_x = nn.Parameter(torch.from_numpy(sobel_x).float().expand(self.classes, 1, 3, 3),
                                    requires_grad=False)
        self.sobel_y = nn.Parameter(torch.from_numpy(sobel_y).float().expand(self.classes, 1, 3, 3),
                                    requires_grad=False)

        self.diff_x = nn.Conv2d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_x.weight = self.sobel_x
        self.diff_y = nn.Conv2d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_y.weight = self.sobel_y

    def forward(self, predication, label):
        grd_x = self.diff_x(predication)
        grd_y = self.diff_y(predication)

        # length
        length = torch.sum(
            torch.abs(torch.sqrt(grd_x ** 2 + grd_y ** 2 + 1e-8)))
        length = (length - length.min()) / (length.max() - length.min() + 1e-8)
        length = torch.sum(length)

        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = self.miu * region_in + region_out

        return region + length


class ACLossV2(nn.Module):
    """
    Active Contour Loss
    based on maxpooling & minpooling
    """

    def __init__(self, miu=1.0, classes=3):
        super(ACLossV2, self).__init__()

        self.miu = miu
        self.classes = classes

    def forward(self, predication, label):
        min_pool_x = nn.functional.max_pool2d(
            predication * -1, (3, 3), 1, 1) * -1
        contour = torch.relu(nn.functional.max_pool2d(
            min_pool_x, (3, 3), 1, 1) - min_pool_x)

        # length
        length = torch.sum(torch.abs(contour))

        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = self.miu * region_in + region_out

        return region + length


def ACELoss(y_pred, y_true, u=1, a=1, b=1):
    """
    Active Contour Loss
    based on total variations and mean curvature
    """
    def first_derivative(input):
        u = input
        m = u.shape[2]
        n = u.shape[3]

        ci_0 = (u[:, :, 1, :] - u[:, :, 0, :]).unsqueeze(2)
        ci_1 = u[:, :, 2:, :] - u[:, :, 0:m - 2, :]
        ci_2 = (u[:, :, -1, :] - u[:, :, m - 2, :]).unsqueeze(2)
        ci = torch.cat([ci_0, ci_1, ci_2], 2) / 2

        cj_0 = (u[:, :, :, 1] - u[:, :, :, 0]).unsqueeze(3)
        cj_1 = u[:, :, :, 2:] - u[:, :, :, 0:n - 2]
        cj_2 = (u[:, :, :, -1] - u[:, :, :, n - 2]).unsqueeze(3)
        cj = torch.cat([cj_0, cj_1, cj_2], 3) / 2

        return ci, cj

    def second_derivative(input, ci, cj):
        u = input
        m = u.shape[2]
        n = u.shape[3]

        cii_0 = (u[:, :, 1, :] + u[:, :, 0, :] -
                 2 * u[:, :, 0, :]).unsqueeze(2)
        cii_1 = u[:, :, 2:, :] + u[:, :, :-2, :] - 2 * u[:, :, 1:-1, :]
        cii_2 = (u[:, :, -1, :] + u[:, :, -2, :] -
                 2 * u[:, :, -1, :]).unsqueeze(2)
        cii = torch.cat([cii_0, cii_1, cii_2], 2)

        cjj_0 = (u[:, :, :, 1] + u[:, :, :, 0] -
                 2 * u[:, :, :, 0]).unsqueeze(3)
        cjj_1 = u[:, :, :, 2:] + u[:, :, :, :-2] - 2 * u[:, :, :, 1:-1]
        cjj_2 = (u[:, :, :, -1] + u[:, :, :, -2] -
                 2 * u[:, :, :, -1]).unsqueeze(3)

        cjj = torch.cat([cjj_0, cjj_1, cjj_2], 3)

        cij_0 = ci[:, :, :, 1:n]
        cij_1 = ci[:, :, :, -1].unsqueeze(3)

        cij_a = torch.cat([cij_0, cij_1], 3)
        cij_2 = ci[:, :, :, 0].unsqueeze(3)
        cij_3 = ci[:, :, :, 0:n - 1]
        cij_b = torch.cat([cij_2, cij_3], 3)
        cij = cij_a - cij_b

        return cii, cjj, cij

    def region(y_pred, y_true, u=1):
        label = y_true.float()
        c_in = torch.ones_like(y_pred)
        c_out = torch.zeros_like(y_pred)
        region_in = torch.abs(torch.sum(y_pred * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - y_pred) * ((label - c_out) ** 2)))
        region = u * region_in + region_out
        return region

    def elastica(input, a=1, b=1):
        ci, cj = first_derivative(input)
        cii, cjj, cij = second_derivative(input, ci, cj)
        beta = 1e-8
        length = torch.sqrt(beta + ci ** 2 + cj ** 2)
        curvature = (beta + ci ** 2) * cjj + \
                    (beta + cj ** 2) * cii - 2 * ci * cj * cij
        curvature = torch.abs(curvature) / ((ci ** 2 + cj ** 2) ** 1.5 + beta)
        elastica = torch.sum((a + b * (curvature ** 2)) * torch.abs(length))
        return elastica

    loss = region(y_pred, y_true, u=u) + elastica(y_pred, a=a, b=b)
    return loss


class ACLoss3D(nn.Module):
    """
    Active Contour Loss
    based on sobel filter
    """

    def __init__(self, classes=4, alpha=1):
        super(ACLoss3D, self).__init__()
        self.alpha = alpha
        self.classes = classes
        sobel = np.array([[[1., 2., 1.],
                           [2., 4., 2.],
                           [1., 2., 1.]],

                          [[0., 0., 0.],
                           [0., 0., 0.],
                           [0., 0., 0.]],

                          [[-1., -2., -1.],
                           [-2., -4., -2.],
                           [-1., -2., -1.]]])

        self.sobel_x = nn.Parameter(
            torch.from_numpy(sobel.transpose(0, 1, 2)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)
        self.sobel_y = nn.Parameter(
            torch.from_numpy(sobel.transpose(1, 0, 2)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)
        self.sobel_z = nn.Parameter(
            torch.from_numpy(sobel.transpose(1, 2, 0)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)

        self.diff_x = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_x.weight = self.sobel_x
        self.diff_y = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_y.weight = self.sobel_y
        self.diff_z = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_z.weight = self.sobel_z

    def forward(self, predication, label):
        grd_x = self.diff_x(predication)
        grd_y = self.diff_y(predication)
        grd_z = self.diff_z(predication)

        # length
        length = torch.sqrt(grd_x ** 2 + grd_y ** 2 + grd_z ** 2 + 1e-8)
        length = (length - length.min()) / (length.max() - length.min() + 1e-8)
        length = torch.sum(length)

        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = region_in + region_out

        return self.alpha * region + length


class ACLoss3DV2(nn.Module):
    """
    Active Contour Loss
    based on minpooling & maxpooling
    """

    def __init__(self, miu=1.0, classes=3):
        super(ACLoss3DV2, self).__init__()

        self.miu = miu
        self.classes = classes

    def forward(self, predication, label):
        min_pool_x = nn.functional.max_pool3d(
            predication * -1, (3, 3, 3), 1, 1) * -1
        contour = torch.relu(nn.functional.max_pool3d(
            min_pool_x, (3, 3, 3), 1, 1) - min_pool_x)

        # length
        length = torch.sum(torch.abs(contour))

        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = self.miu * region_in + region_out

        return region + length


class ACELoss3D(nn.Module):
    """
    Active contour based elastic model loss
    based on total variations and mean curvature
    """

    def __init__(self, alpha=1e-3, beta=1.0, miu=1, classes=3):
        super(ACELoss3D, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.miu = miu

    def first_derivative(self, input):
        u = input
        m = u.shape[2]
        n = u.shape[3]
        k = u.shape[4]

        ci_0 = (u[:, :, 1, :, :] - u[:, :, 0, :, :]).unsqueeze(2)
        ci_1 = u[:, :, 2:, :, :] - u[:, :, 0:m - 2, :, :]
        ci_2 = (u[:, :, -1, :, :] - u[:, :, m - 2, :, :]).unsqueeze(2)
        ci = torch.cat([ci_0, ci_1, ci_2], 2) / 2

        cj_0 = (u[:, :, :, 1, :] - u[:, :, :, 0, :]).unsqueeze(3)
        cj_1 = u[:, :, :, 2:, :] - u[:, :, :, 0:n - 2, :]
        cj_2 = (u[:, :, :, -1, :] - u[:, :, :, n - 2, :]).unsqueeze(3)
        cj = torch.cat([cj_0, cj_1, cj_2], 3) / 2

        ck_0 = (u[:, :, :, :, 1] - u[:, :, :, :, 0]).unsqueeze(4)
        ck_1 = u[:, :, :, :, 2:] - u[:, :, :, :, 0:k - 2]
        ck_2 = (u[:, :, :, :, -1] - u[:, :, :, :, k - 2]).unsqueeze(4)
        ck = torch.cat([ck_0, ck_1, ck_2], 4) / 2

        return ci, cj, ck

    def second_derivative(self, input, ci, cj, ck):
        u = input
        m = u.shape[2]
        n = u.shape[3]
        k = u.shape[4]

        cii_0 = (u[:, :, 1, :, :] + u[:, :, 0, :, :] -
                 2 * u[:, :, 0, :, :]).unsqueeze(2)
        cii_1 = u[:, :, 2:, :, :] + \
            u[:, :, :-2, :, :] - 2 * u[:, :, 1:-1, :, :]
        cii_2 = (u[:, :, -1, :, :] + u[:, :, -2, :, :] -
                 2 * u[:, :, -1, :, :]).unsqueeze(2)
        cii = torch.cat([cii_0, cii_1, cii_2], 2)

        cjj_0 = (u[:, :, :, 1, :] + u[:, :, :, 0, :] -
                 2 * u[:, :, :, 0, :]).unsqueeze(3)
        cjj_1 = u[:, :, :, 2:, :] + \
            u[:, :, :, :-2, :] - 2 * u[:, :, :, 1:-1, :]
        cjj_2 = (u[:, :, :, -1, :] + u[:, :, :, -2, :] -
                 2 * u[:, :, :, -1, :]).unsqueeze(3)

        cjj = torch.cat([cjj_0, cjj_1, cjj_2], 3)

        ckk_0 = (u[:, :, :, :, 1] + u[:, :, :, :, 0] -
                 2 * u[:, :, :, :, 0]).unsqueeze(4)
        ckk_1 = u[:, :, :, :, 2:] + \
            u[:, :, :, :, :-2] - 2 * u[:, :, :, :, 1:-1]
        ckk_2 = (u[:, :, :, :, -1] + u[:, :, :, :, -2] -
                 2 * u[:, :, :, :, -1]).unsqueeze(4)

        ckk = torch.cat([ckk_0, ckk_1, ckk_2], 4)

        cij_0 = ci[:, :, :, 1:n, :]
        cij_1 = ci[:, :, :, -1, :].unsqueeze(3)

        cij_a = torch.cat([cij_0, cij_1], 3)
        cij_2 = ci[:, :, :, 0, :].unsqueeze(3)
        cij_3 = ci[:, :, :, 0:n - 1, :]
        cij_b = torch.cat([cij_2, cij_3], 3)
        cij = cij_a - cij_b

        cik_0 = ci[:, :, :, :, 1:n]
        cik_1 = ci[:, :, :, :, -1].unsqueeze(4)

        cik_a = torch.cat([cik_0, cik_1], 4)
        cik_2 = ci[:, :, :, :, 0].unsqueeze(4)
        cik_3 = ci[:, :, :, :, 0:k - 1]
        cik_b = torch.cat([cik_2, cik_3], 4)
        cik = cik_a - cik_b

        cjk_0 = cj[:, :, :, :, 1:n]
        cjk_1 = cj[:, :, :, :, -1].unsqueeze(4)

        cjk_a = torch.cat([cjk_0, cjk_1], 4)
        cjk_2 = cj[:, :, :, :, 0].unsqueeze(4)
        cjk_3 = cj[:, :, :, :, 0:k - 1]
        cjk_b = torch.cat([cjk_2, cjk_3], 4)
        cjk = cjk_a - cjk_b

        return cii, cjj, ckk, cij, cik, cjk

    def region(self, y_pred, y_true, u=1):
        label = y_true.float()
        c_in = torch.ones_like(y_pred)
        c_out = torch.zeros_like(y_pred)
        region_in = torch.abs(torch.sum(y_pred * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - y_pred) * ((label - c_out) ** 2)))
        region = u * region_in + region_out
        return region

    def elastica(self, input, a=1, b=1):
        ci, cj, ck = self.first_derivative(input)
        cii, cjj, ckk, cij, cik, cjk = self.second_derivative(
            input, ci, cj, ck)
        beta = 1e-8
        length = torch.sqrt(beta + ci ** 2 + cj ** 2 + ck ** 2)
        curvature = (1 + ci ** 2 + cj ** 2) * ckk + (1 + cj ** 2 + ck ** 2) * cii + (
            1 + ci ** 2 + ck ** 2) * cjj - 2 * cik * cjk * cij
        curvature = torch.abs(curvature) / \
            ((1 + ci ** 2 + cj ** 2 + ck ** 2) ** 0.5 + beta)
        elastica = torch.sum(a + b * (curvature ** 2) * torch.abs(length))
        return elastica

    def forward(self, y_pred, y_true):
        loss = self.region(y_pred, y_true, u=self.miu) + \
            self.elastica(y_pred, a=self.alpha, b=self.beta)
        return loss


class FastACELoss3D(nn.Module):
    """
    Active contour based elastic model loss
    based on sobel and laplace filter
    """

    def __init__(self, miu=1, alpha=1e-3, beta=2.0, classes=4, types="laplace"):
        super(FastACELoss3D, self).__init__()
        self.miu = miu
        self.alpha = alpha
        self.beta = beta
        self.classes = classes
        self.types = types
        sobel = np.array([[[1., 2., 1.],
                           [2., 4., 2.],
                           [1., 2., 1.]],

                          [[0., 0., 0.],
                           [0., 0., 0.],
                           [0., 0., 0.]],

                          [[-1., -2., -1.],
                           [-2., -4., -2.],
                           [-1., -2., -1.]]])
        laplace_kernel = np.ones((3, 3, 3))
        laplace_kernel[1, 1, 1] = -26

        self.sobel_x = nn.Parameter(
            torch.from_numpy(sobel.transpose(0, 1, 2)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)
        self.sobel_y = nn.Parameter(
            torch.from_numpy(sobel.transpose(1, 0, 2)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)
        self.sobel_z = nn.Parameter(
            torch.from_numpy(sobel.transpose(1, 2, 0)).float().unsqueeze(0).unsqueeze(0).expand(self.classes, 1, 3, 3,
                                                                                                3), requires_grad=False)
        self.laplace = nn.Parameter(
            torch.from_numpy(laplace_kernel).float().unsqueeze(
                0).unsqueeze(0).expand(self.classes, 1, 3, 3, 3),
            requires_grad=False)

        self.diff_x = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_x.weight = self.sobel_x
        self.diff_y = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_y.weight = self.sobel_y
        self.diff_z = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1, padding=1,
                                bias=False)
        self.diff_z.weight = self.sobel_z

        self.laplace_operator = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1,
                                          padding=1,
                                          bias=False)
        self.laplace_operator.weight = self.laplace

    def forward(self, predication, label):
        grd_x = self.diff_x(predication)
        grd_y = self.diff_y(predication)
        grd_z = self.diff_z(predication)
        diff = self.laplace_operator(predication)

        # length
        length = torch.sqrt(grd_x ** 2 + grd_y ** 2 + grd_z ** 2 + 1e-8)
        length = (length - length.min()) / (length.max() - length.min() + 1e-8)

        # curvature
        if self.types:
            curvature = torch.abs(diff)
            curvature = (curvature - curvature.min()) / \
                (curvature.max() - curvature.min() + 1e-8)
        else:
            """
            maybe more powerful
            """
            curvature = torch.abs(
                diff) / ((grd_x ** 2 + grd_y ** 2 + grd_z ** 2 + 1) ** 0.5 + 1e-8)
            curvature = (curvature - curvature.min()) / \
                (curvature.max() - curvature.min() + 1e-8)
        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = self.miu * region_in + region_out

        # elastic
        elastic = torch.sum((self.alpha + self.beta * curvature ** 2) * length)
        return region + elastic


class FastACELoss3DV2(nn.Module):
    """
    Active contour based elastic model loss
    based on minpooling & maxpooling and laplace filter
    """

    def __init__(self, miu=1, alpha=1e-3, beta=2.0, classes=4, types= None):
        super(FastACELoss3DV2, self).__init__()
        self.miu = miu
        self.alpha = alpha
        self.beta = beta
        self.classes = classes
        self.types = types
        laplace_kernel = np.ones((3, 3, 3))
        laplace_kernel[1, 1, 1] = -26

        self.laplace = nn.Parameter(
            torch.from_numpy(laplace_kernel).float().unsqueeze(
                0).unsqueeze(0).expand(self.classes, 1, 3, 3, 3),
            requires_grad=False)

        self.laplace_operator = nn.Conv3d(self.classes, self.classes, groups=self.classes, kernel_size=3, stride=1,
                                          padding=1,
                                          bias=False)
        self.laplace_operator.weight = self.laplace

    def forward(self, predication, label):
        min_pool_x = nn.functional.max_pool3d(predication * -1, 3, 1, 1) * -1
        contour = torch.relu(nn.functional.max_pool3d(
            min_pool_x, 3, 1, 1) - min_pool_x)

        diff = self.laplace_operator(predication)

        # length
        length = torch.abs(contour)

        # curvature
        if self.types:
            curvature = torch.abs(diff)
            curvature = (curvature - curvature.min()) / \
                (curvature.max() - curvature.min() + 1e-8)
        else:
            """
            maybe more powerful
            """
            curvature = torch.abs(diff) / ((length ** 2 + 1) ** 0.5 + 1e-8)
            curvature = (curvature - curvature.min()) / \
                (curvature.max() - curvature.min() + 1e-8)
        # region
        label = label.float()
        c_in = torch.ones_like(predication)
        c_out = torch.zeros_like(predication)
        region_in = torch.abs(torch.sum(predication * ((label - c_in) ** 2)))
        region_out = torch.abs(
            torch.sum((1 - predication) * ((label - c_out) ** 2)))
        region = self.miu * region_in + region_out

        # elastic
        elastic = torch.sum((self.alpha + self.beta * curvature ** 2) * length)
        return region + elastic